/*
 * Copyright 2007 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

/*
 * A minimal mechanism to make our agent discoverable via UDP
 * broadcast, multicast, or unicast queries.
 */


#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <string.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <pthread.h>
#include <libgen.h>
#include <pwd.h>
#include <grp.h>

#ifdef SMF
#include <libscf.h>
#endif

#include "sthelper.h"
#include "discoverable.h"
#include "scache.h"
#include "scf.h"

#include "strlcpy.h"

/* Size of the buffer we'll use for various purposes. */
#define BUF_LEN (128 * 1024)

/* Max delay for random sleep, in microseconds -- set to 1/10th of a second */
#define MAX_RANDOM_SLEEP 1000

/* Timeout value for network inactivity in inetd invocations */
#define IDLE_MAX (120)

/* The default UDP port on which we'll listen for discovery queries. */
#define DEFAULT_UDP_PORT 6481

/* The default TCP port on which we'll listen for communication requests. */
#define DEFAULT_TCP_PORT 6481

/* Maximum allowable TCP port */
#define MAX_PORT 65535

struct udp_msg
{
  struct sockaddr_storage addr;
  socklen_t               addrlen;
  char content[BUF_LEN];
};

/* 
 * Table of permitted SMF properties that can be used to override
 * default values.  Used for SMF (inetd) control only.
 */

scf_param_t smfprops[] =
{
    { SCF_ST_PROBE_REGISTRY, "registry_location",  NULL },
    { SCF_ST_PROBE_TCPPORT , "discovery_tcp_port", NULL },
    { SCF_ST_PROBE_NULL,     NULL,                 NULL }
};

static void handle_discovery_message(int, struct udp_msg *, int);
static void make_daemon_process();
static void set_uid(char *);
static void set_gid(char *);

static void random_sleep();

/* Service registry tag state information */
st_env_t st_env;

int stdiscover_verbose;

int main(int argc, char **argv)
{
  int udp_port = DEFAULT_UDP_PORT;
  int tcp_port = DEFAULT_TCP_PORT;
  int inetd = 0;
  int probe_fd;
  int do_daemon = 1;
  int errors = 0;
  int c;
  char *registry_location = NULL;
  char *uid = NULL;
  char *gid = NULL;
  extern char *optarg;
  extern int optind;
  extern int opterr;
  extern int optopt;
  int rval;
  int rlen;
  
  while ((c = getopt(argc, argv, "fvir:u:t:N:U:G:l:")) != -1)
  {
    switch (c)
    {
      case 'f':
        do_daemon = 0;
        break;
      case 'i':
        inetd = 1;
        break;
      case 'v':
        stdiscover_verbose++;
        break;
      case 'r':
        rlen = strlen(optarg) + 1; 
        if ((registry_location = (char *)malloc(rlen)) == NULL)
        {
          perror("malloc(registry_location)");
          return 1;
        }
        if (strlcpy(registry_location, optarg, rlen) >= rlen)
        {
          fprintf(stderr, "-r registry path too long\n");
          return 1;
        }
        break;
      case 'u':
        udp_port = atoi(optarg);
        break;
      case 't':
        tcp_port = atoi(optarg);
        break;
      case 'U':
        rlen = strlen(optarg) + 1;
        if ((uid = (char *)malloc(rlen)) == NULL)
        {
          perror("malloc(uid)");
          return 1;
        }
        if (strlcpy(uid, optarg, rlen) >= rlen)
        {
          fprintf(stderr, "-U user id name too long\n");
          return 1;
        }
        break;
      case 'G':
        rlen = strlen(optarg) + 1;
        if ((gid = (char *)malloc(rlen)) == NULL)
        {
          perror("malloc(gid)");
          return 1;
        }
        if (strlcpy(gid, optarg, rlen) >= rlen)
        {
          fprintf(stderr, "-G group id name too long\n");
          return 1;
        }
        break;
      case 'l':
        if ((rval = open(optarg,(O_RDWR|O_APPEND|O_CREAT|O_EXCL),0666)) > 0)
        {
           (void) dup2(rval, 2);
           (void) close(rval);
        }
        else
        {
           /* Non-fatal error */
           fprintf(stderr, "Unable to open log file: %s\n", optarg);
        } 
        break;
      case ':':
        fprintf(stderr, "Option -%c requires an operand\n", optopt);
        errors++;
        break;
      case '?':
        fprintf(stderr, "Unrecognized option: -%c\n", optopt);
        errors++;
        break;
    }
  }

  if (errors)
  {
    fprintf(stderr, "Usage: %s [options]\n", basename(argv[0]));
    fprintf(stderr, "Options:\n");
    fprintf(stderr, "  -f                     do not fork() from the calling shell\n");
    fprintf(stderr, "  -l logfile             redirect stderr to specified file\n");
    fprintf(stderr, "  -i                     run under control of inetd(1M)\n");
    fprintf(stderr, "  -v                     provide verbose output\n");
    fprintf(stderr, "  -r alt_root_dir        look for Service Tag Registry under specified dir\n");
    fprintf(stderr, "  -u udp_port            use the specified udp port for st-disc discovery\n");
    fprintf(stderr, "  -t tcp_port            advertise this TCP port for st-disc discovery\n");
    fprintf(stderr, "  -U user                setuid to specified user\n");
    fprintf(stderr, "  -G group               setgid to specified group\n");

    return errors;
  }

#ifdef SMF
    /* 
     *  In inetd-invocations, the published SMF properties
     *  (if present) override any existing values.
     */
	if (inetd)
    {
        process_smf_properties(smfprops);

        for (int i = 0; smfprops[i].id != SCF_ST_PROBE_NULL; i++)
        {
            int port = 0;
            char *propval = smfprops[i].value;
            if (propval != NULL && strlen(propval) > 0)
            {
                 switch (smfprops[i].id)
                {
                    case SCF_ST_PROBE_REGISTRY:
                        if (registry_location) free(registry_location);
                        rlen = strlen(propval) + 1;
                        if ((registry_location = (char *)malloc(rlen)) != NULL)
                        {
                            strlcpy(registry_location, propval, rlen);
                        }
                        break;
                    case SCF_ST_PROBE_TCPPORT:
                        if (((port = atoi(propval)) != 0 ) &&
                             (port > IPPORT_RESERVED) &&
                             (port <= MAX_PORT))
                        {
                           tcp_port = port;
                        }
                        break;
                    default:
                        break; 
                 }
            }
        }
    }
#endif

  if ( (!inetd) && do_daemon)
  {
     make_daemon_process();
  }

  if (gid != NULL)
  {
    set_gid(gid);
  }
  if (uid != NULL)
  {
    set_uid(uid);
  }

  memset(&st_env, 0, sizeof st_env);
  if (st_initialize(registry_location, &st_env) < 0)
  {
    free(registry_location);
  }

  if (inetd)
  {
    probe_fd = 0;
  }

  receive_discovery_messages(udp_port, tcp_port, inetd, probe_fd);

  return(0);
}

/*
 * Loops indefinitely awaiting a UDP message on the specified port.  When
 * a message is received, it calls handle_discovery_message.  If this 
 * executable was invoked via inetd(1M), the incoming "probe_fd" is the
 * descriptor for the passed-in UDP socket.
 */

void
receive_discovery_messages(int udp_port, int tcp_port, int inetd, int probe_fd)
{
  int sockfd;
#ifndef LINUX
  struct sockaddr_in6 my_addr;
#else
  struct sockaddr_in my_addr;
#endif
  struct udp_msg *msg;
  socklen_t addr_len;
  int numbytes;
  char hostname[MAXHOSTNAMELEN + 1];
  struct sockaddr_storage from;
  socklen_t fromlen;

  if (!inetd)
  {
#ifndef LINUX
     if ((sockfd = socket(AF_INET6, SOCK_DGRAM, 0)) < 0)
#else
     if ((sockfd = socket(AF_INET, SOCK_DGRAM, 0)) < 0)
#endif
     {
       perror("socket");
       exit(1);
     }
     (void) memset(&my_addr, 0, sizeof my_addr);
#ifndef LINUX
     my_addr.sin6_family = AF_INET6;
     my_addr.sin6_port = htons(udp_port);
     my_addr.sin6_addr = in6addr_any;
#else
     my_addr.sin_family = AF_INET;
     my_addr.sin_port = htons(udp_port);
     my_addr.sin_addr.s_addr = INADDR_ANY;
#endif

     if (bind(sockfd, (struct sockaddr *)&my_addr, sizeof(my_addr)) < 0)
     {
       perror("bind (udp)");
       exit(1);
     }
  }
  else
  {
     /* Socket zero is passed via inetd, already opened */
     sockfd = probe_fd;
  }

  /* Get length of incoming address, based on the address family */

  fromlen = sizeof from;
  if (getsockname(sockfd, (struct sockaddr *)&from,
                  &fromlen) < 0) {
      perror("getsockname");
      exit(1);
  }

  switch(from.ss_family) {
      case AF_INET:
         addr_len = (socklen_t)  sizeof(struct sockaddr_in);
         break;
#ifndef LINUX
      case AF_INET6:
         addr_len = (socklen_t)  sizeof(struct sockaddr_in6);
         break;
#endif
      default:
         fprintf(stderr, "Unknown address family: %d\n", 
                 from.ss_family);        
         exit(1);
  }
  
  for (;;)
  {
    msg = (struct udp_msg *)(malloc(sizeof(struct udp_msg)));
    if (msg == NULL)
    {
      perror("malloc(msg)");
      continue;
    }

    /* Wait a fixed amount of time for the next incoming probe */

    fd_set readfds;
    struct timeval timer;

    FD_ZERO(&readfds);
    FD_SET(sockfd, &readfds);
    timer.tv_sec =  IDLE_MAX; 
    timer.tv_usec = 0;
  
    int n = select(sockfd + 1, &readfds, NULL, NULL, &timer);

    if (n < 0) {
       if (errno == EINTR) {
          if (msg) {
             free(msg);
          }
          continue;
       }
       else {
          perror("select");
          exit(1);
       }
    }

    /* Timeout:  if inetd invocation, exit due to inactivity */

    if (n == 0) {
       if (inetd)
          exit(0);
       else {
          if (msg) {
             free(msg);
          }
          continue;
       }
    }

    /* Data is available for reading */

    if ((numbytes = recvfrom(sockfd, msg->content, BUF_LEN-1 , 0,
        (struct sockaddr *)(&(msg->addr)), &addr_len)) < 0)
    {
       perror("recvfrom");
       exit(1);
    }

    msg->addrlen = addr_len;

    msg->content[numbytes] = '\0';
    if (stdiscover_verbose > 0)
    {
       int name_not_found  = getnameinfo((const struct sockaddr *) &(msg->addr),
#ifndef LINUX
                            sizeof(struct sockaddr_in6),
#else
                            sizeof(struct sockaddr_in),
#endif
                            hostname, sizeof hostname,
                            NULL, 0, 0);
       if (name_not_found)
       {
          (void) strlcpy(hostname, "unknown", sizeof(hostname));
       }

       fprintf(stderr, " got %4d bytes from %s: %s\n",
              numbytes, hostname, msg->content);
    }

    /* handle_discovery_message will take care of free'ing msg */
    handle_discovery_message(sockfd, msg, tcp_port);

  }
}

/*
 * Handles a discovery message.
 *
 * @param msg pointer to a malloc'd buf containing the message to be
 *  handled; upon return from this method, will have been freed.
 */
static void
handle_discovery_message(int sockfd, struct udp_msg *msg, int tcp_port)
{
  char send_buf[BUF_LEN];

  /*
   * Only respond to connections which a present a cookie value
   * which is not yet in our cache.  The purpose for this is to
   * allow a stcollector to retransmit requests (since UDP is
   * unreliable) without triggering redundant responses
   * from stagents who got and responded to the original
   * transmission.
   */
  fprintf(stderr, "Registry: %s\n", st_get_registry_urn());
  if (scache_contains(msg->content))
  {
    fprintf(stderr, "NOTICE: ignoring due to scache hit: %s\n", msg->content);
  }
  else
  {
    scache_add(msg->content);

    /* random delay avoids flooding requesters broadcasting to many agents */
    random_sleep();

    fprintf(stderr, "got msg: %s\n", msg->content);
    snprintf(send_buf, BUF_LEN, "%s %d\n", st_get_registry_urn(), tcp_port);
    fprintf(stderr, "udp response: %s\n", send_buf);

    if (sendto(sockfd, send_buf, strlen(send_buf), 0,
        (struct sockaddr *)(&(msg->addr)), msg->addrlen) < 0)
    {
      perror("sendto");
    }
  }

  free(msg);
}


/*
 * Sleep for a random time of between 0 and MAX_RANDOM_SLEEP microseconds.
 */
static void
random_sleep()
{
  unsigned sleep_time;

  srand(time(NULL));
  sleep_time = (random() % MAX_RANDOM_SLEEP);

  fprintf(stderr, "(random sleep: %u microseconds)\n", sleep_time);
  usleep(sleep_time);
}


static void
make_daemon_process()
{
  /* BEGIN - attempt to make this a daemon */
  /* Ignore terminal I/O signals */
  signal(SIGTTOU, SIG_IGN);
  signal(SIGTTIN, SIG_IGN);
  signal(SIGTSTP, SIG_IGN);
  /* Ignore terminal I/O signals */

  int pid;
  pid = fork();
  if (pid != 0) {
    exit(0); /* this is the parent process here, so let's exit it for our daemon work */
  }

  setsid(); /* child process, so let's make us a session leader, and lose control tty */
  signal(SIGHUP, SIG_IGN); /* SIGHUP will be sent to child process so ignore it */

  if (pid != 0) {
    exit(0);
  }
  /* END - attempt to make this a daemon */
}

/* Attempts to change the uid and gid of the program to the svctag user
 * and the svctag's group (which should be svctag) */
static void
set_uid(char *user_name)
{
  struct passwd *user;

  if (user_name == NULL)
  {
    fprintf(stderr, "WARNING: no user specified\n");
    return;
  }

  user = getpwnam(user_name);

  if (user == NULL)
  {
    fprintf(stderr, "WARNING: no such user exists: %s\n", user_name);
    return;
  }

  /* attempt to set the uid and gid of the program -- only works if
   * the user has appropriate permissions */
  if (setuid(user->pw_uid) != 0)
  {
    perror("setuid");
  }
}


/* Attempts to change the uid and gid of the program to the svctag user
 * and the svctag's group (which should be svctag) */
static void
set_gid(char *group_name)
{
  struct group *group;

  if (group_name == NULL)
  {
    fprintf(stderr, "WARNING: no group specified\n");
    return;
  }

  group = getgrnam(group_name);

  if (group == NULL)
  {
    fprintf(stderr, "WARNING: no such group exists: %s\n", group_name);
    return;
  }

  if (setgid(group->gr_gid) != 0)
  {
    perror("setgid");
  }
}
