/*
 *
 * Copyright 2007 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 *
 */

/*
 * Service Tag Helper
 *
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <dirent.h>
#include <unistd.h>
#include <limits.h>
#include <pwd.h>
#include <errno.h>
#include <libxml/encoding.h>
#include <libxml/parser.h>
#include <libxml/tree.h>
#include <libxml/xmlmemory.h>

#include "sthelper.h"

#include "strlcpy.h"
#include "strlcat.h"

#ifdef HAVE_UUID_UUID_H  
#include <uuid/uuid.h>
#else
#include "uuid.h"
#endif

/*
 * Private Function Prototypes 
 *
 */
static  int     find_instance_urn(xmlDocPtr, xmlNodePtr, char *);
static  int     find_product_urn(xmlDocPtr, xmlNodePtr, char *, char *);
static  int     count_records_by_uid(xmlDocPtr, xmlNodePtr, char *);
static  void    get_timestamp(char *);
static  void    str_upper(char *);
static  void    fix_escape_char(char *, char *);
static  int     is_reg_valid(st_env_t *);
static  int     is_authorized(st_t *);

static  char    registry_urn[MAX_URN_LEN];


/*
 * st_is_initialized(st_env_t *env) 
 *
 * Return 1 (true) is environment is initialized else 0 (false)
 *
 */
int
st_is_initialized(st_env_t *env)
{
	return (env->initialized_flag == ST_INITIALIZED) ? ST_INITIALIZED : ST_NOT_INITIALIZED;
}

/*
 * st_initialize(char *root_dir, st_env_t *) 
 *
 * Initialize the service tag environment.  Return ST_ERR_NO_ERROR if environment was initialized 
 * else returns error code.
 *
 */
int
st_initialize(char *root_dir, st_env_t *env)
{
	int        rc;
	xmlDocPtr  doc       = NULL;
	xmlNodePtr root_node = NULL;
	xmlDtdPtr  dtd       = NULL;
	char       dtd_file[PATH_MAX + 1 + MAX_FILE_NAME_LEN];
	DIR        *dirp;
	int        fd;

	if (st_is_initialized(env) == ST_INITIALIZED) {
		return ST_ERR_ALREADY_INIT;
	}
	memset(registry_urn, '\0', sizeof(registry_urn));

	/*
	 * This initializes the library and check potential ABI mismatches
	 * between the version it was compiled for and the actual shared
	 * library used.
	 */
	LIBXML_TEST_VERSION

	/* 
	 * Construct the full file path name and check the service tag directory.
	 * The default Service Tag Registry is /usr/var/sadm/servicetag/registry/servicetag.xml.
	 */
	if (root_dir == NULL || strlen(root_dir) == 0) {
		if (strlcpy(env->root_dir, SERVICE_TAG_DEFAULT_ROOT, PATH_MAX) >= PATH_MAX) 
			return ST_ERR_MAX_LEN_EXCEEDED;
		if (strlcpy(env->registry_path, env->root_dir, PATH_MAX) >= PATH_MAX)
			return ST_ERR_MAX_LEN_EXCEEDED;
		if (strlcat(env->registry_path, SERVICE_TAG_PATH, PATH_MAX) >= PATH_MAX)
			return ST_ERR_MAX_LEN_EXCEEDED;
	} else {
		if (strlcpy(env->root_dir, root_dir, PATH_MAX) >= PATH_MAX)
			return ST_ERR_MAX_LEN_EXCEEDED;
		if (strlcpy(env->registry_path, env->root_dir, PATH_MAX) >= PATH_MAX)
			return ST_ERR_MAX_LEN_EXCEEDED;
                /* Concatenate the file separator only when the path is not "/" to prevent "//" */ 
                if (strcmp(env->registry_path, SERVICE_TAG_DEFAULT_ROOT) != 0) { 
			if (strlcat(env->registry_path, FILE_SEPARATOR, PATH_MAX) >= PATH_MAX)
				return ST_ERR_MAX_LEN_EXCEEDED;
		}
		if (strlcat(env->registry_path, SERVICE_TAG_PATH, PATH_MAX) >= PATH_MAX)
			return ST_ERR_MAX_LEN_EXCEEDED;
	}
	if ((dirp = opendir(env->registry_path)) == NULL) {
		return ST_ERR_NO_DIR;
	} else {
		closedir(dirp);
	}
	snprintf(env->registry, PATH_MAX + 1 + MAX_FILE_NAME_LEN, "%s%s%s", env->registry_path, FILE_SEPARATOR, SERVICE_TAG_FILE);

	/*
	 * Check if the registry exists
	 */ 

	if ((fd = open(env->registry, O_RDWR, S_IRUSR | S_IWUSR | S_IRGRP)) != -1) {
		/*
		 * Registry exists; parse it and validate the DOM
		 */
		close(fd);
		if (is_reg_valid(env) == 1) {
			rc = ST_ERR_NO_ERROR;
		} else {
			rc = ST_ERR_XML_INVALID_REG;
		}
	} else {
		/*
		 * Registry does not exist; create a new Service Tag Registry
		 */
		doc = xmlNewDoc(BAD_CAST "1.0");
		if (doc == NULL) {
			return ST_ERR_XML_NEW_DOC; 
		}
		root_node = xmlNewNode(NULL, BAD_CAST ST_ROOT);
		rc = generate_urn(env->registry_urn, SERVICE_TAG_NID);
                if (rc != ST_ERR_NO_ERROR) {
			xmlFreeDoc(doc);
		} else {
			strlcpy(env->registry_version, SERVICE_TAG_VERSION, MAX_PRODUCT_VERSION_LEN);
			xmlNewProp(root_node, BAD_CAST ST_URN,     BAD_CAST env->registry_urn);
			xmlNewProp(root_node, BAD_CAST ST_VERSION, BAD_CAST env->registry_version);
			xmlDocSetRootElement(doc, root_node);
			/*
			 * Creates a DTD declaration. Isn't mandatory.
			 */
			if (strcmp(env->root_dir, SERVICE_TAG_DEFAULT_ROOT) == 0) {  
				snprintf(dtd_file, PATH_MAX + 1 + MAX_FILE_NAME_LEN,
					"%s%s%s", SERVICE_TAG_DTD_PATH, FILE_SEPARATOR, SERVICE_TAG_DTD);
			} else { 
				snprintf(dtd_file, PATH_MAX + 1 +  MAX_FILE_NAME_LEN, 
					"%s%s%s", env->registry_path, FILE_SEPARATOR, SERVICE_TAG_DTD);
			}
			dtd = xmlCreateIntSubset(doc, BAD_CAST ST_ROOT, NULL, BAD_CAST dtd_file);
			if (dtd == NULL) {
				xmlFreeDoc(doc);
				return ST_ERR_XML_CREATE_SUB;
			}
			rc = xmlSaveFileEnc(env->registry, doc, SERVICE_TAG_ENCODING);
			if (rc == -1) {
				rc = ST_ERR_XML_SAVE_REG;
			} else {
				rc = ST_ERR_NO_ERROR;
			}
            (void) chmod(env->registry, S_IRUSR | S_IWUSR);

			xmlFreeDoc(doc);
			/* xmlFreeDtd(dtd); */
		}
	}
	if (rc == ST_ERR_NO_ERROR) {
		if (strlcpy(registry_urn, env->registry_urn, MAX_URN_LEN) >= MAX_URN_LEN) {
			rc = ST_ERR_MAX_LEN_EXCEEDED;
			env->initialized_flag = ST_NOT_INITIALIZED;
		} else {
			env->initialized_flag = ST_INITIALIZED;
		}
	} else { 
		env->initialized_flag = ST_NOT_INITIALIZED;
	}

	return rc;
}


/*
 * st_close(st_env_t *env)
 *
 * Clean up the libxml environment 
 *
 */
void
st_close(st_env_t *env)
{
	if (st_is_initialized(env) == ST_INITIALIZED) {
		env->initialized_flag = ST_NOT_INITIALIZED;
		xmlCleanupParser();
	}
}


/*
 * st_add(st_t *s, st_env_t *env)
 *
 * Add a service tag in the registry.  Requires product name, version, URN, parent
 * vendor, platform, container and source.   Return ST_ERR_NO_ERROR if service tag
 * was added else return error code.
 *
 */
int
st_add(st_t *s, st_env_t *env)
{
	st_t p;
	xmlDocPtr  doc       = NULL;
	xmlNodePtr root_node = NULL;
	xmlNodePtr st_node   = NULL;
	char uid_str[10];
	int recs             = 0;
	int rc;

	if (st_is_initialized(env) == ST_NOT_INITIALIZED) {
		return ST_ERR_NOT_INIT;
	}

	/* 
	 * Check if instance URN is a duplicate
	 */
	memset(&p, '\0', sizeof(p));
	if (strlcpy(p.instance_urn, s->instance_urn, sizeof(p.instance_urn)) >= sizeof(p.instance_urn))
		return ST_ERR_MAX_LEN_EXCEEDED;
	if (st_get(&p, env) == 0) {
		return ST_ERR_DUP_REC;
	}

	/* 
	 * Validate parameters 
	 */

	if (s->product_name == NULL || strlen(s->product_name) == 0 || 
		s->product_version == NULL || strlen(s->product_version) == 0 || 
		s->product_urn    == NULL || strlen(s->product_urn) == 0    || 
		s->product_parent == NULL || strlen(s->product_parent) == 0 ||
		s->product_vendor == NULL || strlen(s->product_vendor) == 0 ||
		s->platform_arch  == NULL || strlen(s->platform_arch)  == 0 ||
		s->container      == NULL || strlen(s->container)      == 0 ||
		s->source         == NULL || strlen(s->source)         == 0) {
		return ST_ERR_MISSING_FIELD;
	}

	/* 
	 * Check data lengths 
	 */

	if (strlen(s->product_name) > (MAX_PRODUCT_NAME_LEN - 1) || 
		strlen(s->product_version)  > (MAX_PRODUCT_VERSION_LEN - 1) ||
		strlen(s->product_urn)    > (MAX_URN_LEN - 1) ||
		strlen(s->product_parent_urn) > (MAX_URN_LEN -1) ||
		strlen(s->product_parent) > (MAX_PRODUCT_PARENT_LEN - 1) ||
		strlen(s->product_defined_inst_id) > (MAX_PRODUCT_PARENT_LEN - 1) ||
		strlen(s->product_vendor) > (MAX_PRODUCT_VENDOR_LEN - 1)     ||
		strlen(s->platform_arch)  > (MAX_PLATFORM_ARCH_LEN  - 1)     ||
		strlen(s->container)      > (MAX_CONTAINER_LEN -1) ||
		strlen(s->source)         > (MAX_SOURCE_LEN -1)) {
		return ST_ERR_MAX_LEN_EXCEEDED;
	}

	/* Parse the registry and get the DOM */
	doc = xmlParseFile(env->registry);
	if (doc == NULL) {
		return ST_ERR_XML_PARSE;
	}

	/* Get the root element node */
	root_node = xmlDocGetRootElement(doc);
	if (root_node == NULL) {
		xmlFreeDoc(doc);
		return ST_ERR_XML_EMPTY_DOC;
	}

        /* Count the registry records for this user and abort if greater than limit */
	s->installer_uid = getuid();
	snprintf(uid_str, sizeof(uid_str), "%d", s->installer_uid);
	recs = count_records_by_uid(doc, root_node, uid_str);
        if ((recs + 1) > MAX_SERVICE_TAGS_PER_USER) {
		xmlFreeDoc(doc);
		return ST_ERR_MAX_REC_PER_USER_EXCEEDED;
        }
        

	/* Add an element named ST_MAIN as child of registry. */
	st_node = xmlNewNode(NULL, BAD_CAST ST_MAIN);

	/* 
	 * Generate product instance unique identifier if it is not supplied 
         */
	/* Write an element named ST_INSTANCE_URN as child of service_tag. */
	if (s->instance_urn == NULL || strlen(s->instance_urn) == 0) {
		rc = generate_urn(s->instance_urn, SERVICE_TAG_NID);
		if (rc != ST_ERR_NO_ERROR) {
			xmlFreeDoc(doc);
			return rc;
		}
	}
	xmlNewTextChild(st_node, NULL, BAD_CAST ST_INSTANCE_URN, BAD_CAST s->instance_urn);

	xmlNewTextChild(st_node, NULL, BAD_CAST ST_PRODUCT_NAME, BAD_CAST s->product_name);
	xmlNewTextChild(st_node, NULL, BAD_CAST ST_PRODUCT_VERSION, BAD_CAST s->product_version);
	xmlNewTextChild(st_node, NULL, BAD_CAST ST_PRODUCT_URN, BAD_CAST s->product_urn);
	if (s->product_parent_urn == NULL || strlen(s->product_parent_urn) == 0)
		xmlNewTextChild(st_node, NULL, BAD_CAST ST_PRODUCT_PARENT_URN, BAD_CAST "");
	else
		xmlNewTextChild(st_node, NULL, BAD_CAST ST_PRODUCT_PARENT_URN, BAD_CAST s->product_parent_urn);
	xmlNewTextChild(st_node, NULL, BAD_CAST ST_PRODUCT_PARENT, BAD_CAST s->product_parent);
	if (s->product_defined_inst_id == NULL || strlen(s->product_defined_inst_id) == 0)
		xmlNewTextChild(st_node, NULL, BAD_CAST ST_PRODUCT_DEFINED_INST_ID, BAD_CAST "");
	else
		xmlNewTextChild(st_node, NULL, BAD_CAST ST_PRODUCT_DEFINED_INST_ID, BAD_CAST s->product_defined_inst_id);
	xmlNewTextChild(st_node, NULL, BAD_CAST ST_PRODUCT_VENDOR, BAD_CAST s->product_vendor);
	xmlNewTextChild(st_node, NULL, BAD_CAST ST_PLATFORM_ARCH, BAD_CAST s->platform_arch);

	/* Get the current date/time                                    */
	/* Write an element named ST_TIMESTAMP as child of service_tag. */
	get_timestamp(s->timestamp);
	xmlNewTextChild(st_node, NULL, BAD_CAST ST_TIMESTAMP, BAD_CAST s->timestamp);

	xmlNewTextChild(st_node, NULL, BAD_CAST ST_CONTAINER, BAD_CAST s->container);
	xmlNewTextChild(st_node, NULL, BAD_CAST ST_SOURCE, BAD_CAST s->source);

	/* Write an element named ST_INSTALLER_UID as child of service_tag. */
	s->installer_uid = getuid();
	snprintf(uid_str, sizeof(uid_str), "%d", s->installer_uid);
	xmlNewTextChild(st_node, NULL, BAD_CAST ST_INSTALLER_UID, BAD_CAST uid_str);

	/* Add new node to the root node */
	xmlAddChild(root_node, st_node);


        /* Lock the registry before the write */
        int fd;
        if ((fd = open(env->registry, O_RDWR)) == -1) {
		xmlFreeDoc(doc);
                return ST_ERR_FILE_ERR;
        }
        if (lockf(fd, F_LOCK, (off_t) 0) == -1) {
                // This should be unreacheable.
                // The lock will block if it is use and will automatically continue if the lock is released
		xmlFreeDoc(doc);
                return ST_ERR_LOCKF_ERR;
        }

	/* Write record in registry      */

	if (xmlSaveFileEnc(env->registry, doc, SERVICE_TAG_ENCODING) == -1) 
		rc = ST_ERR_XML_SAVE_REG;
	else
		rc = ST_ERR_NO_ERROR;
	xmlFreeDoc(doc);

        /* Unlock the registry before the write */
        if (lockf(fd, F_ULOCK, (off_t) 0) == -1) {
                // This should be unreacheable.
                perror("lockf");
        }
        close(fd);
        
	return rc;
}


/*
 * st_update(st_t *s, st_env_t *env)
 *
 * Update a service tag in the registry.  Requires instance URN as the key
 * and either product parent urn or product defined instance id. 
 * Return ST_ERR_NO_ERROR if service tag was updated else return error code.
 *
 */
int
st_update (st_t *s, st_env_t *env)
{
	st_t       p;
	xmlDocPtr  doc        = NULL;
	xmlNodePtr root_node  = NULL;
	xmlNodePtr curr_node  = NULL;
	xmlNodePtr curr_node2 = NULL;
	xmlChar    *val       = NULL;
	char       uid_str[10];
	int        rc;
	int        found;

	found = 0;

	if (st_is_initialized(env) == ST_NOT_INITIALIZED) {
		return ST_ERR_NOT_INIT;
	}

	/* Validate parameters */
	if ((s->instance_urn == NULL || strlen(s->instance_urn) == 0) || 
		((s->product_parent_urn == NULL || strlen(s->product_parent_urn) == 0) &&
		(s->product_defined_inst_id == NULL || strlen(s->product_defined_inst_id) == 0))) {
		return ST_ERR_MISSING_FIELD;
	}

	/*
	 * Check data lengths
	 */

	if (strlen(s->instance_urn) > (MAX_URN_LEN - 1) || 
		strlen(s->product_parent_urn) > (MAX_URN_LEN -1) ||
		strlen(s->product_defined_inst_id) > (MAX_URN_LEN - 1)) {
		return ST_ERR_MAX_LEN_EXCEEDED;
	}

	/* Check if the product instance URN exists */
	memset(&p, '\0', sizeof(p));
        if (strlcpy(p.instance_urn, s->instance_urn, sizeof(p.instance_urn)) >= sizeof(p.instance_urn))
                return ST_ERR_MAX_LEN_EXCEEDED;

	if (st_get(&p, env) != 0) {
		return ST_ERR_REC_NOT_FOUND;
	}
	
	/* Check if user is authorized to update the record */
	if (is_authorized(&p) == 0) {
		return ST_ERR_NOT_AUTH;
	} 

	/* Parse the registry                  */
	doc = xmlParseFile(env->registry);
	if (doc == NULL) {
		return ST_ERR_XML_PARSE;
	}

	/* Get the root element */
	root_node = xmlDocGetRootElement(doc);
	if (root_node == NULL) {
		xmlFreeDoc(doc);
		return ST_ERR_XML_EMPTY_DOC;
	}

	/* Find the instance URN and when found, update the fields */

	curr_node = root_node->xmlChildrenNode;
	while (curr_node != NULL) {
		if ((!xmlStrcmp(curr_node->name, BAD_CAST ST_MAIN))) {
			if (find_instance_urn(doc, curr_node, s->instance_urn)) {
				found = 1;
				curr_node2 = curr_node->xmlChildrenNode;
				while (curr_node2 != NULL) {
					if (!xmlStrcmp(curr_node2->name, BAD_CAST ST_PRODUCT_NAME)) {
						val = xmlNodeListGetString(doc, curr_node2->xmlChildrenNode, 1);
						if (val != NULL) {
							strlcpy(s->product_name, (char *) val, MAX_PRODUCT_NAME_LEN);
							xmlFree(val);
						} else {
							memset(s->product_name, '\0', sizeof(s->product_name));
 						}
					} else if (!xmlStrcmp(curr_node2->name, BAD_CAST ST_PRODUCT_VERSION)) {
						val = xmlNodeListGetString(doc, curr_node2->xmlChildrenNode, 1);
						if (val != NULL) {
							strlcpy(s->product_version, (char *) val, MAX_PRODUCT_VERSION_LEN);
							xmlFree(val);
						} else {
							memset(s->product_version, '\0', sizeof(s->product_version));
						}
					} else if (!xmlStrcmp(curr_node2->name, BAD_CAST ST_PRODUCT_URN)) {
						val = xmlNodeListGetString(doc, curr_node2->xmlChildrenNode, 1);
						if (val != NULL) {
							strlcpy(s->product_urn, (char *) val, MAX_URN_LEN);
							xmlFree(val);
						} else {
							memset(s->product_urn, '\0', sizeof(s->product_urn));
						}
					} else if (!xmlStrcmp(curr_node2->name, BAD_CAST ST_PRODUCT_PARENT_URN)) {
						if (s->product_parent_urn != NULL && strlen(s->product_parent_urn) > 0) {
							xmlNodeSetContent(curr_node2, BAD_CAST s->product_parent_urn);
						} else {
							val = xmlNodeListGetString(doc, curr_node2->xmlChildrenNode, 1);
							if (val != NULL) {
								strlcpy(s->product_parent_urn, (char *) val, MAX_URN_LEN);
								xmlFree(val);
							} else {
								memset(s->product_parent_urn, '\0', sizeof(s->product_parent_urn));
							}
						}
					} else if (!xmlStrcmp(curr_node2->name, BAD_CAST ST_PRODUCT_PARENT)) {
						val = xmlNodeListGetString(doc, curr_node2->xmlChildrenNode, 1);
						if (val != NULL) {
							strlcpy(s->product_parent, (char *) val, MAX_PRODUCT_PARENT_LEN);
							xmlFree(val);
						} else {
							memset(s->product_parent, '\0', sizeof(s->product_parent));
						}
					} else if (!xmlStrcmp(curr_node2->name, BAD_CAST ST_PRODUCT_DEFINED_INST_ID)) {
						if (s->product_defined_inst_id != NULL && strlen(s->product_defined_inst_id) > 0) {
							xmlNodeSetContent(curr_node2, BAD_CAST s->product_defined_inst_id);
						} else {
							val = xmlNodeListGetString(doc, curr_node2->xmlChildrenNode, 1);
							if (val != NULL) {
								strlcpy(s->product_defined_inst_id, (char *) val, MAX_URN_LEN);
								xmlFree(val);
							} else {
								memset(s->product_defined_inst_id, '\0', sizeof(s->product_defined_inst_id));
							}
						}
					} else if (!xmlStrcmp(curr_node2->name, BAD_CAST ST_PRODUCT_VENDOR)) {
						val = xmlNodeListGetString(doc, curr_node2->xmlChildrenNode, 1);
						if (val != NULL) { 
							strlcpy(s->product_vendor, (char *) val, MAX_PRODUCT_VENDOR_LEN);
							xmlFree(val);
						} else {
							memset(s->product_vendor, '\0', sizeof(s->product_vendor));
						}
					} else if (!xmlStrcmp(curr_node2->name, BAD_CAST ST_PLATFORM_ARCH)) {
						val = xmlNodeListGetString(doc, curr_node2->xmlChildrenNode, 1);
						if (val != NULL) {
							strlcpy(s->platform_arch, (char *) val, MAX_PLATFORM_ARCH_LEN);
							xmlFree(val);
						} else {
							memset(s->platform_arch, '\0', sizeof(s->platform_arch));
						}
					} else if (!xmlStrcmp(curr_node2->name, BAD_CAST ST_CONTAINER)) {
						val = xmlNodeListGetString(doc, curr_node2->xmlChildrenNode, 1);
						if (val != NULL) {
							strlcpy(s->container, (char *) val, MAX_CONTAINER_LEN);
							xmlFree(val);
						} else {
							memset(s->container, '\0', sizeof(s->container));
						}
					} else if (!xmlStrcmp(curr_node2->name, BAD_CAST ST_SOURCE)) {
						val = xmlNodeListGetString(doc, curr_node2->xmlChildrenNode, 1);
						if (val != NULL) {
							strlcpy(s->source, (char *) val, MAX_SOURCE_LEN);
							xmlFree(val);
						} else {
							memset(s->source, '\0', sizeof(s->source));
						}
					} else if (!xmlStrcmp(curr_node2->name, BAD_CAST ST_INSTALLER_UID)) {
						s->installer_uid = getuid();
						snprintf(uid_str, sizeof(uid_str), "%d", s->installer_uid);
						xmlNodeSetContent(curr_node2, BAD_CAST uid_str);
					} else if (!xmlStrcmp(curr_node2->name, BAD_CAST ST_TIMESTAMP)) {
						get_timestamp(s->timestamp);
						xmlNodeSetContent(curr_node2, BAD_CAST s->timestamp);
					}
					curr_node2 = curr_node2->next;
				}
				break;
			}
		}
		curr_node = curr_node->next;
	}
	if (found) {
		/* Lock the registry before the write */
		int fd;
		if ((fd = open(env->registry, O_RDWR)) == -1) {
			xmlFreeDoc(doc);
			return ST_ERR_FILE_ERR;
		}
		if (lockf(fd, F_LOCK, (off_t) 0) == -1) {
			// This should be unreacheable.
			// The lock will block if it is use and will automatically continue if the lock is released
			xmlFreeDoc(doc);
			return ST_ERR_LOCKF_ERR;
		}

		/* Save the registry */
		if (xmlSaveFileEnc(env->registry, doc, SERVICE_TAG_ENCODING) == -1)
			rc = ST_ERR_XML_SAVE_REG;
		else
			rc = ST_ERR_NO_ERROR;

		/* Unlock the registry before the write */
		if (lockf(fd, F_ULOCK, (off_t) 0) == -1) {
			// This should be unreacheable.
			perror("lockf");
		}
		close(fd);

	} else {
		rc = ST_ERR_REC_NOT_FOUND;
	}
	xmlFreeDoc(doc);
	return rc;
}


/*
 * st_get(st_t *s, st_env_t *env)
 *
 * Get a service tag in the registry.  Requires instance URN as the key
 * and return the data in the service tag structure.  Return ST_ERR_NO_ERROR
 * if service tag was found else return the error code.
 *
 */
int
st_get (st_t *s, st_env_t *env)
{
	xmlDocPtr  doc        = NULL;
	xmlNodePtr root_node  = NULL;
	xmlNodePtr curr_node  = NULL;
	xmlNodePtr curr_node2 = NULL;
	xmlChar    *val       = NULL;
	char       save_urn[MAX_URN_LEN];
	int        rc;
	int        found;

	found = 0;

	if (st_is_initialized(env) == ST_NOT_INITIALIZED) {
		return ST_ERR_NOT_INIT;
	}

	/* Validate product instance URN */
	if (s->instance_urn == NULL || strlen(s->instance_urn) == 0) {
		return ST_ERR_MISSING_FIELD;
	}
	if (strlen(s->instance_urn) > MAX_URN_LEN - 1) {
		return ST_ERR_MAX_LEN_EXCEEDED;
	}
	
	/* Clear out the service tag structure */
	if (strlcpy(save_urn, s->instance_urn, sizeof(save_urn)) >= sizeof(save_urn))
		return ST_ERR_MAX_LEN_EXCEEDED;
	memset(s, '\0', sizeof(st_t));
	if (strlcpy(s->instance_urn, save_urn, MAX_URN_LEN) >= MAX_URN_LEN)
		return ST_ERR_MAX_LEN_EXCEEDED;
	
	/* Check if registry is valid          */
	if (is_reg_valid(env) == 0) {
		return ST_ERR_XML_INVALID_REG;
	}

	/* Parse the registry                  */
	doc = xmlParseFile(env->registry);
	if (doc == NULL) {
		return ST_ERR_XML_PARSE;
	}

	/* Get the root element */
	root_node = xmlDocGetRootElement(doc);
	if (root_node == NULL) {
		xmlFreeDoc(doc);
		return ST_ERR_XML_EMPTY_DOC;
	}

	/* Find the instance URN and when found, retrieve the fields */

	curr_node = root_node->xmlChildrenNode;
	while (curr_node != NULL) {
		if ((!xmlStrcmp(curr_node->name, BAD_CAST ST_MAIN))) {
			if (find_instance_urn(doc, curr_node, s->instance_urn)) {
				found = 1;
				curr_node2 = curr_node->xmlChildrenNode;
				while (curr_node2 != NULL) {
					if (!xmlStrcmp(curr_node2->name, BAD_CAST ST_PRODUCT_NAME)) {
						val = xmlNodeListGetString(doc, curr_node2->xmlChildrenNode, 1);
						if (val != NULL) {
							strlcpy(s->product_name, (char *) val, MAX_PRODUCT_NAME_LEN);
							xmlFree(val);
						} else {
							memset(s->product_name, '\0', sizeof(s->product_name));
						}
					} else if (!xmlStrcmp(curr_node2->name, BAD_CAST ST_PRODUCT_VERSION)) {
						val = xmlNodeListGetString(doc, curr_node2->xmlChildrenNode, 1);
						if (val != NULL) {
							strlcpy(s->product_version, (char *) val, MAX_PRODUCT_VERSION_LEN);
							xmlFree(val);
						} else {
							memset(s->product_version, '\0', sizeof(s->product_version));
						}
					} else if (!xmlStrcmp(curr_node2->name, BAD_CAST ST_PRODUCT_URN)) {
						val = xmlNodeListGetString(doc, curr_node2->xmlChildrenNode, 1);
						if (val != NULL) {
							strlcpy(s->product_urn, (char *) val, MAX_URN_LEN);
							xmlFree(val);
						} else {
							memset(s->product_urn, '\0', sizeof(s->product_urn));
						}
					} else if (!xmlStrcmp(curr_node2->name, BAD_CAST ST_PRODUCT_PARENT_URN)) {
						val = xmlNodeListGetString(doc, curr_node2->xmlChildrenNode, 1);
						if (val != NULL) {
							strlcpy(s->product_parent_urn, (char *) val, MAX_URN_LEN);
							xmlFree(val);
						} else {
							memset(s->product_parent_urn, '\0', sizeof(s->product_parent_urn));
						}
					} else if (!xmlStrcmp(curr_node2->name, BAD_CAST ST_PRODUCT_PARENT)) {
						val = xmlNodeListGetString(doc, curr_node2->xmlChildrenNode, 1);
						if (val != NULL) {
							strlcpy(s->product_parent, (char *) val, MAX_PRODUCT_PARENT_LEN);
							xmlFree(val);
						} else {
							memset(s->product_parent, '\0', sizeof(s->product_parent));
						}
					} else if (!xmlStrcmp(curr_node2->name, BAD_CAST ST_PRODUCT_DEFINED_INST_ID)) {
						val = xmlNodeListGetString(doc, curr_node2->xmlChildrenNode, 1);
						if (val != NULL) {
							strlcpy(s->product_defined_inst_id, (char *) val, MAX_URN_LEN);
							xmlFree(val);
						} else {
							memset(s->product_defined_inst_id, '\0', sizeof(s->product_defined_inst_id));
						}
					} else if (!xmlStrcmp(curr_node2->name, BAD_CAST ST_PRODUCT_VENDOR)) {
						val = xmlNodeListGetString(doc, curr_node2->xmlChildrenNode, 1);
						if (val != NULL) {
							strlcpy(s->product_vendor, (char *) val, MAX_PRODUCT_VENDOR_LEN);
							xmlFree(val);
						} else {
							memset(s->product_vendor, '\0', sizeof(s->product_vendor));
						}
					} else if (!xmlStrcmp(curr_node2->name, BAD_CAST ST_PLATFORM_ARCH)) {
						val = xmlNodeListGetString(doc, curr_node2->xmlChildrenNode, 1);
						if (val != NULL) {
							strlcpy(s->platform_arch, (char *) val, MAX_PLATFORM_ARCH_LEN);
							xmlFree(val);
						} else {
							memset(s->platform_arch, '\0', sizeof(s->platform_arch));
						}
					} else if (!xmlStrcmp(curr_node2->name, BAD_CAST ST_TIMESTAMP)) {
						val = xmlNodeListGetString(doc, curr_node2->xmlChildrenNode, 1);
						if (val != NULL) {
							strlcpy(s->timestamp, (char *) val, MAX_TIMESTAMP_LEN);
							xmlFree(val);
						} else {
							memset(s->timestamp, '\0', sizeof(s->timestamp));
						}
					} else if (!xmlStrcmp(curr_node2->name, BAD_CAST ST_CONTAINER)) {
						val = xmlNodeListGetString(doc, curr_node2->xmlChildrenNode, 1);
						if (val != NULL) {
							strlcpy(s->container, (char *) val, MAX_CONTAINER_LEN);
							xmlFree(val);
						} else {
							memset(s->container, '\0', sizeof(s->container));
						}
					} else if (!xmlStrcmp(curr_node2->name, BAD_CAST ST_SOURCE)) {
						val = xmlNodeListGetString(doc, curr_node2->xmlChildrenNode, 1);
						if (val != NULL) {
							strlcpy(s->source, (char *) val, MAX_SOURCE_LEN);
							xmlFree(val);
						} else {
							memset(s->source, '\0', sizeof(s->source));
						}
					} else if (!xmlStrcmp(curr_node2->name, BAD_CAST ST_INSTALLER_UID)) {
						val = xmlNodeListGetString(doc, curr_node2->xmlChildrenNode, 1);
						if (val != NULL) {
							s->installer_uid = atoi((char *) val);
							xmlFree(val);
						} else {
							s->installer_uid = -1;
						}
					}
					curr_node2 = curr_node2->next;
				}
				break;
			}
		}
		curr_node = curr_node->next;
	}
	if (found) {
		rc = ST_ERR_NO_ERROR;
	} else {
		rc = ST_ERR_REC_NOT_FOUND;
	}
	xmlFreeDoc(doc);
	return rc;
}


/*
 * st_delete(char *instance_urn, st_env_t *env)
 *
 * Delete a service tag in the registry.  Requires instance URN as the key.
 * Return ST_ERR_NO_ERROR if service tag was deleted else return error code.
 *
 */
int
st_delete (char *instance_urn, st_env_t *env)
{
	st_t       p;
	xmlDocPtr  doc        = NULL;
	xmlNodePtr root_node  = NULL;
	xmlNodePtr curr_node  = NULL;
	int        rc;
	int        found;

	found = 0;

	if (st_is_initialized(env) == ST_NOT_INITIALIZED) {
		return ST_ERR_NOT_INIT;
	}

	/* Validate product instance URN */
	if (instance_urn == NULL || strlen(instance_urn) == 0) {
		return ST_ERR_MISSING_FIELD;
	}
	if (strlen(instance_urn) > MAX_URN_LEN - 1) {
		return ST_ERR_MAX_LEN_EXCEEDED;
	}

	/* Check if the product instance URN exists */
	memset(&p, '\0', sizeof(p));
	if (strlcpy(p.instance_urn, instance_urn, MAX_URN_LEN) >= MAX_URN_LEN) {
		return ST_ERR_MAX_LEN_EXCEEDED;
	}
	if (st_get(&p, env) != 0) {
		return ST_ERR_REC_NOT_FOUND;
	}

	/* Check if user is authorized to delete the record */
	if (is_authorized(&p) == 0) {
		return ST_ERR_NOT_AUTH;
	}

	/* Parse the registry                  */
	doc = xmlParseFile(env->registry);
	if (doc == NULL) {
		return ST_ERR_XML_PARSE;
	}

	/* Get the root element */
	root_node = xmlDocGetRootElement(doc);
	if (root_node == NULL) {
		xmlFreeDoc(doc);
		return ST_ERR_XML_EMPTY_DOC;
	}

	/* Find the instance URN and when found, delete the node */

	curr_node = root_node->xmlChildrenNode;
	while (curr_node != NULL) {
		if ((!xmlStrcmp(curr_node->name, BAD_CAST ST_MAIN))) {
			if (find_instance_urn(doc, curr_node, instance_urn)) {
				found = 1;
				/* Delete the whole service tag node */
				xmlUnlinkNode(curr_node);
				xmlFreeNode(curr_node);
				break;
			}
		}
		curr_node = curr_node->next;
	}
	if (found) {
		/* Lock the registry before the write */
		int fd;
		if ((fd = open(env->registry, O_RDWR)) == -1) {
			xmlFreeDoc(doc);
			return ST_ERR_FILE_ERR;
		}
		if (lockf(fd, F_LOCK, (off_t) 0) == -1) {
			// This should be unreacheable.
			// The lock will block if it is use and will automatically continue if the lock is released
			xmlFreeDoc(doc);
			return ST_ERR_LOCKF_ERR;
		}
                
		if (xmlSaveFileEnc(env->registry, doc, SERVICE_TAG_ENCODING) == -1)
			rc = ST_ERR_XML_SAVE_REG;
		else
			rc = ST_ERR_NO_ERROR;

		/* Unlock the registry before the write */
		if (lockf(fd, F_ULOCK, (off_t) 0) == -1) {
			// This should be unreacheable.
			perror("lockf");
		}
		close(fd);
	
	} else {
		rc = ST_ERR_REC_NOT_FOUND;
	}
	xmlFreeDoc(doc);
	return rc;
}


/*
 * find_instance_urn(xmlDocPtr doc, xmlNodePtr curr_node, char *instance_urn)
 *
 * Find the product instance URN and return 1 if found otherwise return 0.
 *
 */
static int
find_instance_urn(xmlDocPtr doc, xmlNodePtr curr_node, char *instance_urn)
{
	int found = 0;
	xmlChar    *val;
	xmlNodePtr node_ptr = NULL;
	node_ptr = curr_node->xmlChildrenNode;
	while (node_ptr != NULL) {
		if ((!xmlStrcmp(node_ptr->name, BAD_CAST ST_INSTANCE_URN))) {
			val = xmlNodeListGetString(doc, node_ptr->xmlChildrenNode, 1);
			if (val != NULL) {
				if (strcmp((char *) val, instance_urn) == 0)
					found = 1;
				xmlFree(val);
			}
		}
		node_ptr = node_ptr->next;
	}
	return found;
}


/*
 * count_records_by_uid(xmlDocPtr doc, xmlNodePtr curr_node, char *uid)
 *
 * Count the product instance URN in the registry.
 *
 */
static int
count_records_by_uid(xmlDocPtr doc, xmlNodePtr root_node, char *uid)
{
	int count = 0;
	xmlChar    *val;
        xmlNodePtr node_ptr  = NULL;
	xmlNodePtr curr_node = NULL;
	curr_node = root_node->xmlChildrenNode;
	while (curr_node != NULL) {
		if ((!xmlStrcmp(curr_node->name, BAD_CAST ST_MAIN))) {
			node_ptr = curr_node->xmlChildrenNode;
			while (node_ptr != NULL) {
				if ((!xmlStrcmp(node_ptr->name, BAD_CAST ST_INSTALLER_UID))) {
					val = xmlNodeListGetString(doc, node_ptr->xmlChildrenNode, 1);
					if (val != NULL) {
						if (strcmp((char *) val, uid) == 0)
							count++;
						xmlFree(val);
					}
				}
				node_ptr = node_ptr->next; 
			}
		}
		curr_node = curr_node->next;
	}
	return count;
}


/*
 * st_get_instance_urns(char *product_urn, char ***instance_urn_pp, st_env_t *env)
 *
 * Return the count and a pointer to a list of product instance URNs for a 
 * given product URN.  If there is an error, return the error code instead of
 * the count.  Error codes are < 0;
 *
 */
int
st_get_instance_urns(char *product_urn, char ***instance_urn_pp, st_env_t *env)
{
	xmlDocPtr  doc       = NULL;
	xmlNodePtr root_node = NULL;
	xmlNodePtr curr_node = NULL;
	int        i;
	int        found;
	char       work_urn[MAX_URN_LEN];
	char       **instance_urn;

	found = 0;
	*instance_urn_pp = NULL;

	if (st_is_initialized(env) == ST_NOT_INITIALIZED) {
		return ST_ERR_NOT_INIT;
	}

	/*
         * Validate parameter
         */

	if (product_urn == NULL || strlen(product_urn) == 0) {
		return ST_ERR_MISSING_FIELD;
	}
	if (strlen(product_urn) > (MAX_PRODUCT_NAME_LEN - 1)) {
		return ST_ERR_MAX_LEN_EXCEEDED;
        }

	/* Allocate memory for the list of product instance URNs */
	instance_urn = (char **)malloc(sizeof(char *) * MAX_RET_SERVICE_TAGS);
	for (i = 0; i < MAX_RET_SERVICE_TAGS; i++) {
		instance_urn[i] = (char *)malloc(MAX_URN_LEN);  
		if (instance_urn[i] == NULL) {
			return ST_ERR_NO_MEM;
		}
	}

	/* Check if registry is valid          */
	if (is_reg_valid(env) == 0) {
		return ST_ERR_XML_INVALID_REG;
	}

	/* Parse the registry                  */
	doc = xmlParseFile(env->registry);
	if (doc == NULL) {
		return ST_ERR_XML_PARSE;
	}

	/* Get the root element */
	root_node = xmlDocGetRootElement(doc);
	if (root_node == NULL) {
		xmlFreeDoc(doc);
		return ST_ERR_XML_EMPTY_DOC;
	}

        /* Loop through the elements and find the product URN */
	curr_node = root_node->xmlChildrenNode;
	while (curr_node != NULL) {
		if ((!xmlStrcmp(curr_node->name, BAD_CAST ST_MAIN))) {
			if (find_product_urn(doc, curr_node, product_urn, work_urn)) {
				if (found < MAX_RET_SERVICE_TAGS) {
					if (strlcpy(instance_urn[found], work_urn, MAX_URN_LEN) >= MAX_URN_LEN) {
						xmlFreeDoc(doc);
						return ST_ERR_MAX_LEN_EXCEEDED;
					} else {
						found++;
					}
				} else {
					xmlFreeDoc(doc);
					return ST_ERR_MAX_REC_EXCEEDED;
				}
			}
		}
		curr_node = curr_node->next;
	}
	xmlFreeDoc(doc);
	*instance_urn_pp = instance_urn;
	return found;
}

/*
 * st_free_instance_urns(char *instance_urn[]) 
 *
 * Free allocated URN memory
 */
void
st_free_instance_urns(char *instance_urn[])
{
	int i;
	if (instance_urn != NULL) {
		for (i = 0; i < MAX_RET_SERVICE_TAGS; i++) {
			if (instance_urn[i] != NULL) 
				free(instance_urn[i]);
		}
		free(instance_urn);
	}
}


/*
 * find_product_urn(xmlDocPtr doc, xmlNodePtr curr_node, 
 *                  char *product_urn, char *instance_urn)
 *
 * Find the product URN in the given node.  Return 1 if found otherwise return 0.
 * Also return the instance URN of the given product URN.
 *
 */
static int
find_product_urn(xmlDocPtr doc, xmlNodePtr curr_node,
	char *product_urn, char *instance_urn)
{
	int found = 0;
	xmlChar    *val;
	xmlNodePtr node_ptr = NULL;

	node_ptr = curr_node->xmlChildrenNode;
	while (node_ptr != NULL) {
		if ((!xmlStrcmp(node_ptr->name, BAD_CAST ST_PRODUCT_URN))) {
			val = xmlNodeListGetString(doc, node_ptr->xmlChildrenNode, 1);
			if (val != NULL) {
				if (strcmp((char *) val, product_urn) == 0) {
					found = 1;
				}
				xmlFree(val);
			}
		} else if ((!xmlStrcmp(node_ptr->name, BAD_CAST ST_INSTANCE_URN))) {
			val = xmlNodeListGetString(doc, node_ptr->xmlChildrenNode, 1);
			if (val != NULL) {
				strlcpy(instance_urn, (char *) val, MAX_URN_LEN);
				xmlFree(val);
			} else {
				strlcpy(instance_urn, "\0", MAX_URN_LEN);
			}
		} 
		node_ptr = node_ptr->next;
	}
	/* Nullify the instance URN if product URN is not found */
	if (found == 0) 
		strlcpy(instance_urn, "\0", MAX_URN_LEN); 
	return found;
}


/*
 * st_get_registry_doc(doc *xml_doc, st_env_t *env)
 *
 * Return the service tag registry in DOM format.  The return code is set to ST_ERR_NO_ERROR 
 * if successful otherwise it is set to error code.
 *
 */
int
st_get_registry_doc(xmlDocPtr *xml_doc, st_env_t *env)
{
	xmlDocPtr  doc           = NULL;
	xmlDocPtr  new_doc       = NULL;
	xmlNodePtr root_node     = NULL;
	xmlNodePtr new_root_node = NULL;
	xmlNodePtr curr_node     = NULL;
	xmlNodePtr new_node      = NULL; 

	if (st_is_initialized(env) == ST_NOT_INITIALIZED) {
		return ST_ERR_NOT_INIT;
	}

	if (is_reg_valid(env) == 0) {
		return ST_ERR_XML_INVALID_REG;
	}

 	/* Create new XML document             */
	new_doc = xmlNewDoc(BAD_CAST "1.0");
	if (new_doc == NULL) {
		return ST_ERR_XML_NEW_DOC;
	}
	new_root_node = xmlNewNode(NULL, BAD_CAST ST_ROOT);
	if (new_root_node == NULL) {
		xmlFreeDoc(new_doc);
		return ST_ERR_XML_NEW_NODE;
	}
	xmlNewProp(new_root_node, BAD_CAST ST_URN, BAD_CAST env->registry_urn);
	xmlNewProp(new_root_node, BAD_CAST ST_VERSION, BAD_CAST env->registry_version);

	/* Set the service tag version element */
	xmlDocSetRootElement(new_doc, new_root_node);

	/* Parse the registry                  */
 	doc = xmlParseFile(env->registry);
	if (doc == NULL) {
		xmlFreeDoc(new_doc);
		return ST_ERR_XML_PARSE;
	}

	/* Get the root element */
	root_node = xmlDocGetRootElement(doc);
	if (root_node == NULL) {
		xmlFreeDoc(doc);
		xmlFreeDoc(new_doc);
		return ST_ERR_XML_EMPTY_DOC;
	}

	/* Loop through the doc and add the service tag element into the new document */
	curr_node = root_node->xmlChildrenNode;
	while (curr_node != NULL) {
		if ((!xmlStrcmp(curr_node->name, BAD_CAST ST_MAIN))) {
			new_node = xmlCopyNode(curr_node, 1);
			if (new_node != NULL) {
				xmlAddChild(new_root_node, new_node);
			}
		}
		curr_node = curr_node->next;
	}
	xmlFreeDoc(doc);
	*xml_doc = new_doc;
        return ST_ERR_NO_ERROR;
}


/*
 * st_get_registry_xml(char **xml_data, st_env_t *env)
 *
 * Return the service tag registry in XML format.  The return code is set to ST_ERR_NO_ERROR
 * if successful otherwise it is set to the error code.
 *
 */
int
st_get_registry_xml(char **xml_data, st_env_t *env)
{
	xmlDocPtr  doc;
	int        xml_len;
	xmlChar    *xml_buffer;
	int        rc;

	if ((rc = st_get_registry_doc(&doc, env)) != ST_ERR_NO_ERROR) {
		if (doc != NULL) 
			xmlFree(doc);
		return rc;
	}

	/* Generate the text XML from the document */
	xmlDocDumpFormatMemoryEnc(doc, &xml_buffer, &xml_len, SERVICE_TAG_ENCODING, 1);
        *xml_data = (char *) malloc(xml_len + 1);
	if (*xml_data == NULL) {
		xmlFreeDoc(doc);
		xmlFree(xml_buffer);
		return ST_ERR_NO_MEM; 
	}
	if (strlcpy(*xml_data, (char *) xml_buffer, xml_len + 1) >= (xml_len + 1)) {
		xmlFreeDoc(doc);
		xmlFree(xml_buffer);
        	return ST_ERR_MAX_LEN_EXCEEDED;
	} else {
		xmlFreeDoc(doc);
		xmlFree(xml_buffer);
        	return ST_ERR_NO_ERROR;
	}
}

/*
 * st_free_xml(char *xml_data)
 *
 * Free memory allocated by st_get_xml_registry.
 *
 */
void
st_free_xml(char *xml_data) 
{
	free(xml_data);
}


/*
 * generate_urn(char *generated_urn, char *nid) 
 *
 * Generate a unique identifier
 *
 */
int
generate_urn(char* generated_urn, char *nid)
{
	uuid_t  uuid;
	char    uuid_str[MAX_URN_LEN];
#ifdef HAVE_UUID_UUID_H
	uuid_generate(uuid);
        uuid_unparse(uuid, uuid_str);
#else 
	if (uuid_create(&uuid) == 0) {
		return ST_ERR_GEN_UUID;
	}
	uuid_tostring(&uuid, uuid_str);
#endif
	snprintf(generated_urn, MAX_URN_LEN, "urn:%s:%s", nid, uuid_str);
	return ST_ERR_NO_ERROR;
}


/* 
 * get_timestamp(char *gmt) 
 *
 * Return the current date/time in GMT
 *
 */
static void
get_timestamp(char *gmt) 
{
	time_t cal_time;
	struct tm *gmt_time;

	/* Get the current time in GMT */
	time(&cal_time);
	gmt_time = gmtime(&cal_time);

	sprintf(gmt, "%4d-%02d-%02d %02d:%02d:%02d GMT",
		gmt_time->tm_year + 1900, gmt_time->tm_mon + 1, gmt_time->tm_mday,
		gmt_time->tm_hour, gmt_time->tm_min, gmt_time->tm_sec);
}


/*
 * str_upper(char *data)
 *
 * Convert the string to upper case
 *
 */
static void
str_upper(char *data)
{
	int i = 0;
	for (i = 0; i < strlen(data); i++) {
		data[i] = toupper(data[i]);
	}
}


/*
 * fix_escape_char(char *indata, char *outdata)
 *
 * Convert escape char & to &amp;.  xmlSetNodeContent handles <>'"
 * but for some reason does not handle &.  Message received is 
 * "unterminated entity reference".
 */
static void
fix_escape_char(char *indata, char *outdata)
{
	int i;
	int j;
	for (i = j = 0; i < strlen(indata); i++) {
		switch (indata[i]) {
			case '&':
				outdata[j++] = '&';
				outdata[j++] = 'a';
				outdata[j++] = 'm';
				outdata[j++] = 'p';
				outdata[j++] = ';';
				break;
			default:
				outdata[j++] = indata[i];
		}
	}
}


/*
 * is_reg_valid(st_env_t *env)
 *
 * Check the registry against a DTD and return 1 (true) 
 * if valid otherwise return 0 (false)
 *
 */
static int
is_reg_valid(st_env_t *env)
{
        xmlChar          *urn;
        xmlChar          *ver;
	xmlDocPtr        doc;
	xmlNodePtr       root_node;
	xmlValidCtxt     cvp;
	int              rc;

	cvp.userData = (void *) stderr;
	cvp.error    = (xmlValidityErrorFunc) fprintf;
	cvp.warning  = (xmlValidityWarningFunc) fprintf;

	doc = xmlParseFile(env->registry);
	if (doc == NULL) {
		rc = 0;
	} else {
		rc = xmlValidateDocument(&cvp, doc);
		if (rc == 1) {
			root_node = xmlDocGetRootElement(doc);
        		if (root_node == NULL) {
				rc = 0;
			} else {
				urn = xmlGetProp(root_node, BAD_CAST ST_URN);
				ver = xmlGetProp(root_node, BAD_CAST ST_VERSION);
				strlcpy(env->registry_urn, (char *) urn, MAX_URN_LEN);
				strlcpy(env->registry_version, (char *) ver, MAX_PRODUCT_VERSION_LEN);
			}
		}
		xmlFreeDoc(doc);
	}
	return rc;
}


/*
 * is_authorized(st_t *s)
 *
 * Check if user is authorized to update or delete the record.
 * Authorized users are root, svctag and service tag owner/installer.
 * Return 1 if user is authorized otherwise return 0.
 *
 */
static int
is_authorized(st_t *s) 
{
	uid_t uid;
	uid_t svctag_uid;
	struct passwd *p;

	uid = getuid();

	/* Get the uid of the user svctag */
	p = getpwnam(SERVICE_TAG_USER);   
	if (p == NULL)
		svctag_uid = 0;          /* Set to root if not found */
	else
		svctag_uid = p->pw_uid;

	if (uid == 0 || uid == svctag_uid || uid == s->installer_uid) 
		return 1;
	else
		return 0;
}

/*
 * st_get_registry_urn()
 *
 * Return the registry URN.
 *
 */
char *
st_get_registry_urn()
{
        return registry_urn;
}


/*
 * st_error(int st_err_no)
 *
 * Return the human readable error message from st_err_no.
 *
 */
char *
st_error(int st_err_no) 
{
	switch (st_err_no) {
		case ST_ERR_NO_ERROR:
			return "";
			break;
		case ST_ERR_ALREADY_INIT:
			return "Environment already initialized";
			break;
		case ST_ERR_NOT_INIT:
			return "Environment not initialized";
			break;
		case ST_ERR_GEN_UUID:
			return "Unable to generate UUID";
			break;
		case ST_ERR_NO_MEM:
			return "Unable to allocate memory";
			break;
		case ST_ERR_NO_DIR:
			return "Unable to open directory or file";
			break;
		case ST_ERR_NOT_AUTH:
			return "Not authorized";
			break;
		case ST_ERR_XML_PARSE:
			return "Unable to parse XML";
			break;
		case ST_ERR_XML_EMPTY_DOC:
			return "XML file is empty";
			break;
		case ST_ERR_XML_NEW_DOC:
			return "xmlNewDoc failed";
			break;
		case ST_ERR_XML_NEW_NODE:
			return "xmlNewNode failed";
			break;
		case ST_ERR_XML_CREATE_SUB:
			return "xmlCreateIntSubset failed";
			break;
		case ST_ERR_XML_SAVE_REG:
			return "xmlSaveFileEnc failed; unable to save registry";
			break;
		case ST_ERR_XML_INVALID_REG:
			return "Invalid XML";
			break;
		case ST_ERR_DUP_REC:
			return "Duplicate record";
			break;
		case ST_ERR_REC_NOT_FOUND:
			return "Record not found";
			break;
		case ST_ERR_MISSING_FIELD:
			return "Missing required field(s)";
			break;
		case ST_ERR_MAX_LEN_EXCEEDED:
			return "Max. field length exceeded";
			break;
		case ST_ERR_MAX_REC_EXCEEDED:
			return "Max. number of records exceeded";
			break;
		case ST_ERR_FILE_ERR:
			return "File error";
			break;
		case ST_ERR_LOCKF_ERR:
			return "File lock error";
			break;
		case ST_ERR_MAX_REC_PER_USER_EXCEEDED:
			return "Max. number of records per user exceeded";
 			break;
		default:
			return "Unknown error";
			break;
	}
}
/*
 * st_debug(st_env_t *env, st_t *s)
 *
 * Dump the service tag environment and structure
 *
 */
void
st_debug(st_env_t *env, st_t *s)
{
	if (env != NULL) {
		printf("root_dir=%s\n",        env->root_dir);
		printf("registry_path=%s\n",   env->registry_path);
		printf("registry=%s\n",        env->registry);
		printf("registry_urn=%s\n",    env->registry_urn);
		printf("registry_version=%s\n",env->registry_version);
		printf("initialized_flag=%d\n",env->initialized_flag);
	}
	if (s != NULL) {
		printf("instance_urn=%s\n",    s->instance_urn);
		printf("product_name=%s\n",    s->product_name);
		printf("product_version=%s\n", s->product_version);
		printf("product_urn=%s\n",     s->product_urn);
		printf("product_parent_urn=%s\n",  s->product_parent_urn);
		printf("product_parent=%s\n",  s->product_parent);
		printf("product_defined_inst_id=%s\n",  s->product_defined_inst_id);
		printf("product_vendor=%s\n",  s->product_vendor);
		printf("platform_arch=%s\n",   s->platform_arch);
		printf("timestamp=%s\n",       s->timestamp);
		printf("container=%s\n",       s->container);
		printf("source=%s\n",          s->source);
		printf("installer_uid=%d\n",   s->installer_uid);
	}
}
