/*
 *
 * Copyright 2007 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 *
 */

/*
 * Public header file for clients of the Service Tag Registry.
 */

#ifndef _STHELPER_H
#define _STHELPER_H

#pragma ident   "@(#)sthelper.h    1.0   12/05/06 SMI"

#ifdef __cplusplus
extern "C" {
#endif

#include <sys/param.h>
#include <dirent.h>
#include <limits.h>
#include <libxml/encoding.h>
#include <libxml/parser.h>
#include <libxml/tree.h>

#define SERVICE_TAG_VERSION      "1.1.4"
#define SERVICE_TAG_ENCODING     "UTF-8"
#define SERVICE_TAG_DEFAULT_ROOT "/"
#define SERVICE_TAG_PATH         "var/sadm/servicetag/registry"
#define SERVICE_TAG_FILE         "servicetag.xml"
#ifdef LINUX
#define SERVICE_TAG_DTD_PATH     "/opt/sun/servicetag/share/xml/dtd"
#else
#define SERVICE_TAG_DTD_PATH     "/usr/share/lib/xml/dtd"
#endif /* LINUX */
#define SERVICE_TAG_DTD          "servicetag.dtd"
#define SERVICE_TAG_NID          "st"
#define SERVICE_TAG_USER         "svctag"
#define FILE_SEPARATOR           "/"
#ifdef SOLARIS
	#define MAX_FILE_NAME_LEN MAXNAMELEN
#endif
#ifdef LINUX
	#define MAX_FILE_NAME_LEN MAXNAMLEN
#endif


/*
 * Registry Field Lengths
 */

#define MAX_URN_LEN                  256
#define MAX_PRODUCT_NAME_LEN         256
#define MAX_PRODUCT_VERSION_LEN       64
#define MAX_PRODUCT_PARENT_LEN       256
#define MAX_PRODUCT_VENDOR_LEN        64
#define MAX_PLATFORM_ARCH_LEN         64
#define MAX_TIMESTAMP_LEN             25
#define MAX_CONTAINER_LEN             64
#define MAX_SOURCE_LEN                64

#define MAX_RET_SERVICE_TAGS         256 
#define MAX_SERVICE_TAGS_PER_USER    500

/*
 * Registry XML Elements
 */

#define ST_ROOT                      "registry"
#define ST_URN		             "urn"
#define ST_VERSION                   "version"
#define ST_REGISTRY_URN              "registry_urn"
#define ST_MAIN                      "service_tag"
#define ST_INSTANCE_URN              "instance_urn"
#define ST_PRODUCT_NAME              "product_name"
#define ST_PRODUCT_VERSION           "product_version"
#define ST_PRODUCT_URN               "product_urn"
#define ST_PRODUCT_PARENT_URN        "product_parent_urn"
#define ST_PRODUCT_PARENT            "product_parent"
#define ST_PRODUCT_DEFINED_INST_ID   "product_defined_inst_id"
#define ST_PRODUCT_VENDOR            "product_vendor"
#define ST_PLATFORM_ARCH             "platform_arch"
#define ST_TIMESTAMP                 "timestamp"
#define ST_CONTAINER                 "container"
#define ST_SOURCE                "source"
#define ST_INSTALLER_UID         "installer_uid"

/*
 * Structures
 */

typedef struct st_env_struct {
	char root_dir[PATH_MAX];
	char registry_path[PATH_MAX];
	char registry[PATH_MAX + MAX_FILE_NAME_LEN];
	char registry_urn[MAX_URN_LEN];
	char registry_version[MAX_PRODUCT_VERSION_LEN];
	int  initialized_flag;
} st_env_t;

typedef struct st_struct {
	char instance_urn[MAX_URN_LEN];
	char product_name[MAX_PRODUCT_NAME_LEN];
	char product_version[MAX_PRODUCT_VERSION_LEN];
	char product_urn[MAX_URN_LEN];
	char product_parent_urn[MAX_URN_LEN];
	char product_parent[MAX_PRODUCT_PARENT_LEN];
	char product_defined_inst_id[MAX_URN_LEN];
	char product_vendor[MAX_PRODUCT_VENDOR_LEN];
	char platform_arch[MAX_PLATFORM_ARCH_LEN];
	char timestamp[MAX_TIMESTAMP_LEN];
	char container[MAX_CONTAINER_LEN];
	char source[MAX_SOURCE_LEN];
	int  installer_uid;
} st_t;


typedef enum {
	ST_NOT_INITIALIZED = 0,
	ST_INITIALIZED     = 1	
} st_init_level;


typedef enum {
	ST_ERR_NO_ERROR           = 0,
	ST_ERR_ALREADY_INIT       = -32000,
	ST_ERR_NOT_INIT           = -32001,
	ST_ERR_GEN_UUID           = -32002, 
	ST_ERR_NO_MEM             = -32005,
	ST_ERR_NO_DIR             = -32010,
	ST_ERR_NOT_AUTH           = -32011,
	ST_ERR_XML_PARSE          = -32020,
	ST_ERR_XML_EMPTY_DOC      = -32021,
	ST_ERR_XML_NEW_DOC        = -32022,
	ST_ERR_XML_NEW_NODE       = -32023,
	ST_ERR_XML_CREATE_SUB     = -32024,
	ST_ERR_XML_SAVE_REG       = -32025,
	ST_ERR_XML_INVALID_REG    = -32026,
	ST_ERR_DUP_REC            = -32030,
	ST_ERR_REC_NOT_FOUND      = -32031,
	ST_ERR_MISSING_FIELD      = -32032,
	ST_ERR_MAX_LEN_EXCEEDED   = -32033,
	ST_ERR_MAX_REC_EXCEEDED   = -32035,
        ST_ERR_FILE_ERR           = -32050,
        ST_ERR_LOCKF_ERR          = -32051,
	ST_ERR_MAX_REC_PER_USER_EXCEEDED = -32052
} st_error_codes;


/*
 * Function Prototypes
 */

int     st_initialize(char *, st_env_t *);
void    st_close(st_env_t *);
int     st_is_initialized(st_env_t *);
int     st_add(st_t *, st_env_t *);
int     st_update (st_t *, st_env_t *);
int     st_get(st_t *, st_env_t *);
int     st_delete(char *, st_env_t *);
int     st_get_registry_xml(char **, st_env_t *);
void    st_free_xml(char *);
int     st_get_registry_doc(xmlDocPtr *, st_env_t *);
int     st_get_instance_urns(char *, char ***, st_env_t *);
void    st_free_instance_urns(char *[]);
char   *st_get_registry_urn();
char   *st_error(int);
void    st_debug(st_env_t *, st_t *); 
int     generate_urn(char *, char *);

#ifdef  __cplusplus
}
#endif

#endif /* _STHELPER_H */
