/*
 *
 * Copyright 2007 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 *
 */

#pragma ident  "$Id: uuid.c,v 1.1.1.1 2007/04/09 21:22:48 tw76360 Exp $ SMI"

#include <sys/types.h>
#include <netinet/in.h>
#include <inttypes.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <time.h>
#include <fcntl.h>

#if defined(sun)
/* using solaris threads */
#include <thread.h>
#include <synch.h>

#elif defined(__linux__)
#include <pthread.h>
#endif

#include "sysdep.h"
#include "uuid.h"

/* various forward declarations */

static int read_state(unsigned16 *clockseq, uuid_time_t *timestamp,
		      uuid_node_t * node);
static void write_state(unsigned16 clockseq, uuid_time_t timestamp,
			uuid_node_t node);
static void format_uuid_v1(uuid_t * uuid, unsigned16 clockseq,
			   uuid_time_t timestamp, uuid_node_t node);
static void get_current_time(uuid_time_t * timestamp);
static unsigned16 true_random(void);

/* uuid_create -- generate a UUID */

#if defined(sun)
static mutex_t uuid_mutex = DEFAULTMUTEX;
#elif defined(__linux__)
static pthread_mutex_t uuid_mutex;
#endif

int uuid_create(uuid_t * uuid) {

  uuid_time_t timestamp, last_time;
   unsigned16 clockseq;
  uuid_node_t node;
  uuid_node_t last_node;
  int f;

  /* acquire system wide lock so we're alone */

#if defined(sun)
  mutex_lock(&uuid_mutex);
#elif defined(__linux__)
  pthread_mutex_lock(&uuid_mutex);
#endif

  /* get current time */

  get_current_time(&timestamp);

  /* get node ID */

  get_ieee_node_identifier(&node);

  /* get saved state from NV storage */

  f = read_state(&clockseq, &last_time, &last_node);

  /* if no NV state, or if clock went backwards, or node ID changed
     (e.g., net card swap) change clockseq */

  if (!f || memcmp(&node, &last_node, sizeof(uuid_node_t))) {
    if ((clockseq = true_random()) == 0)
       return(0);
  } else if (timestamp < last_time)
    clockseq++;

  /* stuff fields into the UUID */

  format_uuid_v1(uuid, clockseq, timestamp, node);

  /* save the state for next time */

  write_state(clockseq, timestamp, node);

#if defined(sun)
  mutex_unlock(&uuid_mutex);
#elif defined(__linux__)
  pthread_mutex_unlock(&uuid_mutex);
#endif

  return(1);

}

/**
 * Format a UUID as a string into a caller-supplied buffer.
 * The buffer is assumed to be at least UUID_CHARLEN in length.
 */
void uuid_tostring(uuid_t *uuid, char *buf) {

  /*
   * !!!IF YOU CHANGE THIS FORMAT STRING, BE SURE YOU MAKE CORRESPONDING
   * CHANGES TO uuid_fromstring() FUNCTION BELOW!!!
   */
  sprintf(buf,
	  "%8.8x-%4.4x-%4.4x-%2.2x%2.2x-%2.2x%2.2x%2.2x%2.2x%2.2x%2.2x",
	  uuid->time_low,
	  uuid->time_mid,
	  uuid->time_hi_and_version,
	  uuid->clock_seq_hi_and_reserved,
	  uuid->clock_seq_low,
	  uuid->node[0],
	  uuid->node[1],
	  uuid->node[2],
	  uuid->node[3],
	  uuid->node[4],
	  uuid->node[5]);
}

/**
 * Initialize the memory at "uuid" as a valid UUID
 * from the given input string.  The input string is
 * assumed to match the format used in uuid_tostring() above.
 * If that format string changes, this function will need a re-write!!
 *
 * Returns 0 on success, non-zero on failure.
 * In a failure case, the memory at "uuid" is
 * unmodified.
 */
int
uuid_fromstring(uuid_t *uuid, char *inbuf)
{
	uuid_t tmp;
	char workbuf[20];
	char *p = inbuf;
	int i;
	unsigned int scanbuf;

	if (strlen(inbuf) > UUID_CHARLEN) {
		/* XXX: invalid? */
		return (1);
	}

	/* first 8 bytes are "time_low" */
	memcpy(workbuf, p, 8);
	workbuf[8] = '\0';
	if (sscanf(workbuf, "%x", &scanbuf) != 1) {
		return (1);
	}
	tmp.time_low = scanbuf;
	p += 8;

	if (*p++ != '-') {
		/* unexpected format? */
		return (1);
	}

	/* next 4 bytes are "time_mid" */
	memcpy(workbuf, p, 4);
	workbuf[4] = '\0';
	if (sscanf(workbuf, "%x", &scanbuf) != 1) {
		return (1);
	}
	p += 4;
	tmp.time_mid = scanbuf;

	if (*p++ != '-') {
		/* unexpected format? */
		return (1);
	}

	/* next 4 are "time_hi_and_version" */
	memcpy(workbuf, p, 4);
	workbuf[4] = '\0';

	if (sscanf(workbuf, "%x", &scanbuf) != 1) {
		return (1);
	}
	tmp.time_hi_and_version = scanbuf;
	p += 4;

	if (*p++ != '-') {
		/* unexpected format? */
		return (1);
	}

	/* next 2 are "clock_seq_hi_and_reserved" */
	memcpy(workbuf, p, 2);
	workbuf[2] = '\0';
	if (sscanf(workbuf, "%x", &scanbuf) != 1) {
		return (1);
	}
	tmp.clock_seq_hi_and_reserved = scanbuf;

	p += 2;

	/* next 2 are "clock_seq_low" */
	memcpy(workbuf, p, 2);
	workbuf[2] = '\0';
	if (sscanf(workbuf, "%x", &scanbuf) != 1) {
		return (1);
	}
	tmp.clock_seq_low = scanbuf;
	p += 2;

	if (*p++ != '-') {
		return (1);
	}

	/* now we have six pairs of characters that are the hex */
	/* bytes for node[0-5] */
	for (i = 0; i < 6; i++) {
		memcpy(workbuf, p, 2);
		workbuf[2] = '\0';
		if (sscanf(workbuf, "%x", &scanbuf) != 1) {
			return (1);
		}
		tmp.node[i] = scanbuf;
		p += 2;
	}

	memcpy(uuid, &tmp, sizeof (tmp));
	return (0);
}


/* format_uuid_v1 -- make a UUID from the timestamp, clockseq,
   and node ID */

static
void format_uuid_v1(uuid_t * uuid, unsigned16 clock_seq, uuid_time_t
		    timestamp, uuid_node_t node) {
  /* Construct a version 1 uuid with the information we've gathered
   * plus a few constants. */
  uuid->time_low = (unsigned long)(timestamp & 0xFFFFFFFF);
  uuid->time_mid = (unsigned short)((timestamp >> 32) & 0xFFFF);
  uuid->time_hi_and_version = (unsigned short)((timestamp >> 48) &
					       0x0FFF);
  uuid->time_hi_and_version |= (1 << 12);
  uuid->clock_seq_low = clock_seq & 0xFF;
  uuid->clock_seq_hi_and_reserved = (clock_seq & 0x3F00) >> 8;
  uuid->clock_seq_hi_and_reserved |= 0x80;
  memcpy(&uuid->node, &node, sizeof uuid->node);
}

/* data type for UUID generator persistent state */

typedef struct {
  uuid_time_t ts;       /* saved timestamp */
  uuid_node_t node;     /* saved node ID */
  unsigned16 cs;        /* saved clock sequence */
  int inited;
} uuid_state;

static uuid_state st;

static
int read_state(unsigned16 *clockseq,
	       uuid_time_t *timestamp,
	       uuid_node_t *node) {

  if (st.inited) {
    *clockseq = st.cs;
    *timestamp = st.ts;
    *node = st.node;
    return(1);
  }

  return(0);

}

static
void write_state(unsigned16 clockseq,
		 uuid_time_t timestamp,
		 uuid_node_t node) {

  st.cs = clockseq;
  st.ts = timestamp;
  st.node = node;
  st.inited = 1;

}

/* get-current_time -- get time as 60 bit 100ns ticks since whenever.
   Compensate for the fact that real clock resolution is
   less than 100ns. */
static
void get_current_time(uuid_time_t * timestamp) {
  uuid_time_t                time_now;
  static uuid_time_t  time_last;
  static unsigned16   uuids_this_tick;
  static int                   inited = 0;

  if (!inited) {
    get_system_time(&time_now);
    uuids_this_tick = UUIDS_PER_TICK;
    inited = 1;
  };

  for (;;) {
    get_system_time(&time_now);

    /* if clock reading changed since last UUID generated... */
    if (time_last != time_now) {
      /* reset count of uuids gen'd with this clock reading */
      uuids_this_tick = 0;
      break;
    };
    if (uuids_this_tick < UUIDS_PER_TICK) {
      uuids_this_tick++;
      break;
    };
    /* going too fast for our clock; spin */
  };
  /* add the count of uuids to low order bits of the clock reading */
  *timestamp = time_now + uuids_this_tick;
}

/* true_random -- generate a suitable crypto-quality random number. */
static unsigned16
true_random(void)
{
  static int inited = 0;
  uuid_time_t time_now;
  unsigned16  rval;
  int         randfd;

  /* 
   * Attempt to get a suitable quality random number from kernel interface.
   * urandom is used so that we don't block.  Fall back on default
   * method if kernel interface is unavailable.
   */

  if ((randfd = open("/dev/urandom", O_RDONLY)) == -1) {
     if (!inited) {
       get_system_time(&time_now);
       time_now = time_now/UUIDS_PER_TICK;
       srand((unsigned int)(((time_now >> 32) ^ time_now)&0xffffffff));
       inited = 1;
     };

     return (rand());
  }
  else {
     if ((read(randfd, &rval, sizeof rval)) == -1)
        rval = 0;	/* callers had better be checking the retval */
     close(randfd);
     return rval;
  }
}

/* uuid_compare --  Compare two UUID's "lexically" and return
   -1   u1 is lexically before u2
   0   u1 is equal to u2
   1   u1 is lexically after u2
*/
int uuid_compare(uuid_t *u1, uuid_t *u2)
{
  int i;

#define CHECK(f1, f2) if (f1 != f2) return f1 < f2 ? -1 : 1;
  CHECK(u1->time_low, u2->time_low);
  CHECK(u1->time_mid, u2->time_mid);
  CHECK(u1->time_hi_and_version, u2->time_hi_and_version);
  CHECK(u1->clock_seq_hi_and_reserved, u2->clock_seq_hi_and_reserved);
  CHECK(u1->clock_seq_low, u2->clock_seq_low)
    for (i = 0; i < 6; i++) {
      if (u1->node[i] < u2->node[i])
	return -1;
      if (u1->node[i] > u2->node[i])
	return 1;
    }
  return 0;
}
