/*
 * Copyright 2007 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

/*
 * Handles all non-discovery communication with collectors.
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <strings.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/wait.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <signal.h>
#include <pthread.h>
#include <libxml/encoding.h>
#include <libxml/parser.h>
#include <libxml/tree.h>
#include <libxml/xmlmemory.h>
#include <poll.h>
#include <syslog.h>

#ifdef STDEMO
#include <sys/stat.h>  /* needed only for demo */
#endif

#ifdef STCRYPTO
#include "base64.h"

#ifdef STCRYPTO_EVP
#include <openssl/evp.h>
#include "base64.h"
#endif

#ifdef STCRYPTO_UEF
#include <security/cryptoki.h>
#include <security/pkcs11.h>
static CK_BBOOL truevalue = TRUE;
static CK_BBOOL falsevalue = FALSE;
static CK_OBJECT_CLASS class = CKO_SECRET_KEY;
static CK_KEY_TYPE keyType = CKK_AES;
#endif

#endif

#include "sthelper.h"
#include "comm.h"
#include "probe.h"

#include "strlcat.h"
#include "strlcpy.h"

/* Size of the buffer we'll use for various purposes. */
#define BUF_LEN (16 * 1024)
#define DO_ENCRYPT 1
#define DO_NOT_ENCRYPT 0

/*static void handle_get(int, st_env_t *);*/
static void handle_http_get_agent(int);
static void handle_http_get_agent_version(int);
static void handle_http_get_svctag(int, st_env_t *, char *);
static void handle_http_get_svctag_list(int, st_env_t *);
static void handle_http_get_svctag_list_xsl(int);
static void handle_request(int, st_env_t *, char *);

#ifdef STDEMO
static void handle_demo_update(int, st_env_t *);
static void handle_demo_init(int);
static void exec_demo_script(char *, char *, char *);
#endif

#ifdef STCRYPTO_EVP
static int bf_cbc_encrypt(unsigned char *, unsigned const char *, int, unsigned char *);
static int aes_evp_encrypt(unsigned char *, unsigned const char *, int, unsigned char *);
#endif

#ifdef STCRYPTO_UEF
static int aes_uef_encrypt(unsigned char *, unsigned const char *, int, unsigned char *);
#endif

static void encode_chars(const char *, char *, int);

static int sock_printf(int, int, char *, ...);
static char *sock_gets(int, char *, int);
static char *sock_ntop(const struct sockaddr *, socklen_t);
#ifdef LINUX
static void read_and_discard_data(int sockfd);
#endif 

static void get_agent_xml(char *);

extern int stlisten_verbose;

extern char stlisten_version[];

extern char registry_version[];

extern char encryption_key[];
extern int do_encrypt;

/*
 * Listens on INADDR_ANY:port.
 */
void
receive_tcp_connections(int inetd, int local_subnet_only,
                        int local_loopback_only,
                        int port, st_env_t *st_env)
{
  int listen_sockfd;
  int accepted_sockfd;
  socklen_t addr_len;
  struct sockaddr_in my_addr;
  struct sockaddr_in their_addr;
  char buf[BUF_LEN];
  int opt;
  int r;

  struct sockaddr *cliaddr;
  socklen_t len;

  openlog ("Service Tag network responder", LOG_PID | LOG_NDELAY, LOG_LOCAL1);

  if (inetd)
  {
    if ((cliaddr = malloc(128)) == NULL)
    {
      if (stlisten_verbose > 0)
      {
        perror("malloc");
      }
    }

    len = 128;

    getpeername(0, cliaddr, &len);

    if (local_subnet_only)
    {
      opt = 0;
      r = setsockopt(0, IPPROTO_IP, IP_TTL, &opt, sizeof (opt));
      /*
      opt = 0;
      r = setsockopt(0, IPPROTO_IPV6, IPV6_UNICAST_HOPS, &opt, sizeof (opt));
      */
    }

    syslog (LOG_NOTICE|LOG_DAEMON, "Connection from IP address: %s",
      sock_ntop(cliaddr, len));

    if (local_loopback_only)
    {
      if (strncmp("127.0.0.1", sock_ntop(cliaddr, len), 9) != 0) {
        syslog (LOG_NOTICE|LOG_DAEMON, "Rejecting connection (not loopback) from IP address: %s",
            sock_ntop(cliaddr, len));
        free(cliaddr);

        close(0);
        exit(0);
      }
    }

    closelog ();

    sock_gets(0, buf, BUF_LEN);
    handle_request(0, st_env, buf);
#ifdef LINUX
    read_and_discard_data(0);
#endif

    free(cliaddr);

    close(0);
    exit(0);
  }
  else
  {
    if ((listen_sockfd = socket(PF_INET, SOCK_STREAM, 0)) < 0)
    {
      if (stlisten_verbose > 0)
      {
        perror("socket");
      }
      exit(1);
    }

    opt = 1;
    r = setsockopt(listen_sockfd, SOL_SOCKET, SO_REUSEADDR, &opt, sizeof (opt));
    if (r < 0)
    {
      if (stlisten_verbose > 0)
      {
        perror("setsockopt");
      }
      exit(1);
    }

    if (local_subnet_only)
    {
      opt = 0;
      r = setsockopt(listen_sockfd, IPPROTO_IP, IP_TTL, &opt, sizeof (opt));
      /*
      opt = 0;
      r = setsockopt(listen_sockfd, IPPROTO_IPV6, IPV6_UNICAST_HOPS, &opt, sizeof (opt));
      */
    }

    my_addr.sin_family = AF_INET;
    my_addr.sin_port = htons(port);

    if (local_loopback_only)
    {
        inet_pton(AF_INET,"127.0.0.1", &my_addr.sin_addr);
    }
    else
    {
        my_addr.sin_addr.s_addr = INADDR_ANY;
    }

    memset(&(my_addr.sin_zero), '\0', 8);

    if (bind(listen_sockfd, (struct sockaddr *) &my_addr, sizeof(my_addr)) < 0)
    {
      if (stlisten_verbose > 0)
      {
        perror("bind (tcp)");
      }
      exit(1);
    }

    listen(listen_sockfd,5);
    addr_len = sizeof(their_addr);

    for (;;)
    {
      accepted_sockfd
        = accept(listen_sockfd, (struct sockaddr *) &their_addr, &addr_len);

      if (accepted_sockfd < 0)
      {
        if (stlisten_verbose > 0)
        {
          perror("accept");
        }
        continue;
      }

      if (stlisten_verbose > 0)
      {
        fprintf(stderr, "got TCP connection: (fd=%d)\n", accepted_sockfd);
      }

      syslog (LOG_NOTICE|LOG_DAEMON, "Connection from IP address: %s",
        sock_ntop((struct sockaddr*) &their_addr, addr_len));

      closelog ();

      sock_gets(accepted_sockfd, buf, BUF_LEN);

      handle_request(accepted_sockfd, st_env, buf);

#ifdef LINUX
      read_and_discard_data(accepted_sockfd);
#endif LINUX

      close(accepted_sockfd);
    }
  }
}

static void
handle_request(int sockfd, st_env_t *st_env, char *buf)
{
    if (stlisten_verbose > 1)
    {
      fprintf(stderr, "got TCP msg: %s\n", buf);
    }

    if (strncmp(buf, "GET /stv1/agent/version/ HTTP/",
              sizeof("GET /stv1/agent/version/ HTTP/")-1) == 0)
    {
      if (stlisten_verbose > 0)
      {
        fprintf(stderr, "GET agent version\n");
      }
      handle_http_get_agent_version(sockfd);
    }
    else if (strncmp(buf, "GET /stv1/agent/ HTTP/",
              sizeof("GET /stv1/agent/ HTTP/")-1) == 0)
    {
      if (stlisten_verbose > 0)
      {
        fprintf(stderr, "GET agent\n");
      }
      handle_http_get_agent(sockfd);
    }
    else if (strncmp(buf, "GET /stv1/svctag/urn:st:",
              sizeof("GET /stv1/svctag/urn:st:")-1) == 0)
    {
      if (stlisten_verbose > 0)
      {
        fprintf(stderr, "GET svctag\n");
      }
      handle_http_get_svctag(sockfd, st_env, buf);
    }
    else if (strncmp(buf, "GET /stv1/svctag/ HTTP/",
              sizeof("GET /stv1/svctag/ HTTP/")-1) == 0)
    {
      if (stlisten_verbose > 0)
      {
        fprintf(stderr, "GET svctag list\n");
      }
      handle_http_get_svctag_list(sockfd, st_env);
    }
    else if (strncmp(buf, "GET /stv1/st_list.xsl HTTP/",
              sizeof("GET /stv1/st_list.xsl HTTP/")-1) == 0)
    {
      if (stlisten_verbose > 0)
      {
        fprintf(stderr, "GET st_list.xsl\n");
      }
      handle_http_get_svctag_list_xsl(sockfd);
    }
    /*
    else if (strncmp(buf, "[GET]", sizeof("[GET]")-1) == 0)
    {
      if (stlisten_verbose > 0)
      {
        fprintf(stderr, "calling handle_get\n");
      }
      handle_get(sockfd, st_env);
    }
    */
#ifdef STDEMO
    /* BEGIN code for demo */
    else if (strncmp(buf, "[DEMO_INIT]", sizeof("[DEMO_INIT]")-1) == 0)
    {
      if (stlisten_verbose > 0)
      {
        fprintf(stderr, "calling handle_demo_init\n");
      }
      handle_demo_init(sockfd);
    }
    else if (strncmp(buf, "[DEMO_UPDATE]", sizeof("[DEMO_UPDATE]")-1) == 0)
    {
      if (stlisten_verbose > 0)
      {
        fprintf(stderr, "calling handle_demo_update\n");
      }
      handle_demo_update(sockfd, st_env);
    }
    /* END code for demo */
#endif
}


#ifdef STCRYPTO_UEF

static int aes_uef_encrypt(unsigned char *passphrase,
                       unsigned const char *in,
                       int len,
                       unsigned char *out)
{
    CK_RV rv;
    CK_MECHANISM mechanism;
    CK_OBJECT_HANDLE hKey;
    CK_SESSION_HANDLE hSession;
    CK_ULONG ciphertext_len = 64, lastpart_len = 64;
    long ciphertext_space = BUF_LEN;
    CK_ULONG total_encrypted = 0;
    unsigned char padded_buf[BUF_LEN];

    int i, rounds;
    uchar_t *pciphertext;
    uchar_t aes_key[16];
    uchar_t aes_cbc_iv[] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
    int passlen = strlen((char*) passphrase);
    for (i=0; i<passlen; i++) {
        aes_key[i] = passphrase[i];
    }
    for (i=passlen; i<16; i++) {
        aes_key[i] = '\0';
    }

    /* Set the key object */
    CK_ATTRIBUTE template[] = {
        {CKA_CLASS, &class, sizeof (class) },
        {CKA_KEY_TYPE, &keyType, sizeof (keyType) },
        {CKA_TOKEN, &falsevalue, sizeof (falsevalue) },
        {CKA_ENCRYPT, &truevalue, sizeof (truevalue) },
        {CKA_VALUE, &aes_key, sizeof (aes_key) }
    };

    /* Set the encryption mechanism to CKM_AES_CBC_PAD */
    //mechanism.mechanism = CKM_AES_CBC_PAD;
    mechanism.mechanism = CKM_AES_CBC;
    mechanism.pParameter = aes_cbc_iv;
    mechanism.ulParameterLen = 16;

    /* Use SUNW convenience function to initialize the cryptoki
     * library, and open a session with a slot that supports
     * the mechanism we plan on using. */
    rv = SUNW_C_GetMechSession(mechanism.mechanism, &hSession);

    if (rv != CKR_OK) {
        if (stlisten_verbose > 0) {
            fprintf(stderr, "SUNW_C_GetMechSession: rv = 0x%.8X\n", rv);
        }
        exit(1);
    }

    /* Create an object handle for the key */
    rv = C_CreateObject(hSession, template,
        sizeof (template) / sizeof (CK_ATTRIBUTE),
        &hKey);

    if (rv != CKR_OK) {
        if (stlisten_verbose > 0) {
            fprintf(stderr, "C_CreateObject: rv = 0x%.8X\n", rv);
        }
        (void) C_CloseSession(hSession);
        return -5;
    }


    /* Initialize the encryption operation in the session */
    rv = C_EncryptInit(hSession, &mechanism, hKey);

    if (rv != CKR_OK) {
        if (stlisten_verbose > 0) {
            fprintf(stderr, "C_EncryptInit: rv = 0x%.8X\n", rv);
        }
        (void) C_CloseSession(hSession);
        return -4;
    }

    strlcpy((char*) padded_buf, (const char*) in, BUF_LEN);
    if (len % 16 != 0) {
        len = strlen((char*) padded_buf);
        rounds = 16 - (len % 16);
        for (i=0; i<rounds; i++) {
            strlcat((char*) padded_buf, " ", BUF_LEN);
        }
        len = strlen((char*) padded_buf);
    }

    pciphertext = &out[0];
    ciphertext_len = ciphertext_space;
    rv = C_EncryptUpdate(hSession, (CK_BYTE_PTR)padded_buf, len,
        pciphertext, &ciphertext_len);
    pciphertext += ciphertext_len;
    total_encrypted += ciphertext_len;
    ciphertext_space -= ciphertext_len;
    if (rv != CKR_OK) {
        if (stlisten_verbose > 0) {
            fprintf(stderr, "C_EncryptUpdate: rv = 0x%.8X\n", rv);
        }
        return -3;
    }

    /* Get the last portion of the encrypted data */
    lastpart_len = ciphertext_space;
    rv = C_EncryptFinal(hSession, pciphertext, &lastpart_len);
    if (rv != CKR_OK) {
        if (stlisten_verbose > 0) {
            fprintf(stderr, "C_EncryptFinal: rv = 0x%.8X\n", rv);
        }
        return -2;
    }
    total_encrypted += lastpart_len;

    (void) C_CloseSession(hSession);

    return total_encrypted;
}
#endif

#ifdef STCRYPTO_EVP
/* takes as input a passphrase and cleartext input */
/* sets out to the encrypted result */
static int aes_evp_encrypt(unsigned char *passphrase,
                       unsigned const char *in,
                       int len,
                       unsigned char *out)
{
    const EVP_CIPHER *c = EVP_aes_128_cbc();
    EVP_CIPHER_CTX ctx;
    int outl,outl2;
    int i, rounds;
    unsigned char iv[] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
    unsigned char padded_buf[BUF_LEN];

    strlcpy((char*) padded_buf, (const char*) in, BUF_LEN);
    len = strlen((char*) in);

    if (len % 16 != 0) {
        len = strlen((char*) padded_buf);
        rounds = 16 - (len % 16);
        for (i=0; i<rounds; i++) {
            strlcat((char*) padded_buf, " ", BUF_LEN);
        }
        len = strlen((char*) padded_buf);
    }

    EVP_CIPHER_CTX_init(&ctx);
    if(!EVP_EncryptInit_ex(&ctx,c,NULL,passphrase,iv)) {
        if (stlisten_verbose > 0) {
            fprintf(stderr,"EncryptInit failed\n");
        }
        return -1;
    }
    EVP_CIPHER_CTX_set_padding(&ctx, 0);

    if(!EVP_EncryptUpdate(&ctx,out,&outl,padded_buf,len)) {
        if (stlisten_verbose > 0) {
            fprintf(stderr,"Encrypt failed\n");
        }
        return -1;
    }

    if(!EVP_EncryptFinal_ex(&ctx,out+outl,&outl2)) {
        if (stlisten_verbose > 0) {
            fprintf(stderr,"EncryptFinal failed\n");
        }
        return -1;
    }

    EVP_CIPHER_CTX_cleanup(&ctx);
    return outl;
}

/* takes as input a passphrase and cleartext input */
/* sets out to the encrypted result */
static int bf_cbc_encrypt(unsigned char *passphrase,
                          unsigned const char *in,
                          int len,
                          unsigned char *out)
{
    const EVP_CIPHER *c = EVP_bf_cbc();
    EVP_CIPHER_CTX ctx;
    int outl,outl2;
    int i, rounds;
    unsigned char iv[] = {0,0,0,0,0,0,0,0};
    unsigned char padded_buf[BUF_LEN];


    strlcpy((char*) padded_buf, (const char*) in, BUF_LEN);
    len = strlen((char*) in);

    if (len % 16 != 0) {
        len = strlen((char*) padded_buf);
        rounds = 16 - (len % 16);
        for (i=0; i<rounds; i++) {
            strlcat((char*) padded_buf, " ", BUF_LEN);
        }
        len = strlen((char*) padded_buf);
    }

    EVP_CIPHER_CTX_init(&ctx);
    if(!EVP_EncryptInit_ex(&ctx,c,NULL,passphrase,iv)) {
        if (stlisten_verbose > 0) {
            fprintf(stderr,"EncryptInit failed\n");
        }
        return -1;
    }
    EVP_CIPHER_CTX_set_padding(&ctx,0);

    if(!EVP_EncryptUpdate(&ctx,out,&outl,padded_buf,len)) {
        if (stlisten_verbose > 0) {
            fprintf(stderr,"Encrypt failed\n");
        }
        return -1;
    }

    if(!EVP_EncryptFinal_ex(&ctx,out+outl,&outl2)) {
        if (stlisten_verbose > 0) {
            fprintf(stderr,"EncryptFinal failed\n");
        }
        return -1;
    }

    EVP_CIPHER_CTX_cleanup(&ctx);
    return outl;
}
#endif


#ifdef STDEMO

/* BEGIN code for demo */
static void
handle_demo_update(int sockfd, st_env_t *st_env)
{
  /* note that this method only exists for demo purposes
   * it will be removed before moving to production
   */
  struct st_struct s;

  /* Sun Java System Application Server Enterprise Edition 7 2004Q2 */
  char **instUUIDs;
  int rc;

  if ((st_get_instance_urns("urn:uuid:ce95f0e5-6ef7-11d8-9e3a-080020a9ed93",
                            &instUUIDs, st_env) > 0)
   && (instUUIDs != NULL && instUUIDs[0] != NULL && strlen(instUUIDs[0]) > 0))
  {
    if (stlisten_verbose > 0)
    {
      fprintf(stderr, "Found a match for AppServer 7: %s\n", instUUIDs[0]);
    }
    memset(&s,  '\0', sizeof(s));
    strlcpy(s.instance_urn, instUUIDs[0],MAX_URN_LEN);
    strlcpy(s.product_urn, "urn:uuid:ce95f0e5-6ef7-11d8-9e3a-080020a9ed93",MAX_URN_LEN);
    if (stlisten_verbose > 0)
    {
      fprintf(stderr,"\"Uninstalling\" Appserver 7\n");
    }
    rc = st_update(&s, st_env);
    if (stlisten_verbose > 0)
    {
      fprintf(stderr, "return code from update: %d\n", rc);
    }

    /* TODO: should make sure that we haven't already installed
     *  a copy of app server 8.1 */
    memset(&s,  '\0', sizeof(s));
    strlcpy(s.product_name,
      "Sun Java System Application Server Enterprise Edition 8.1 2005Q2",MAX_PRODUCT_NAME_LEN);
    strlcpy(s.product_version, "8.1 2005Q2",MAX_PRODUCT_VERSION_LEN);
    strlcpy(s.product_urn, "urn:uuid:06be2230-f576-11d9-89e6-080020a9ed93",MAX_URN_LEN);
    strlcpy(s.product_parent, "FIXME",MAX_PRODUCT_PARENT_LEN);
    strlcpy(s.product_vendor, "FIXME",MAX_PRODUCT_VENDOR_LEN);
    strlcpy(s.container, "FIXME",MAX_CONTAINER_LEN);
    strlcpy(s.platform_arch, "FIXME",MAX_PLATFORM_ARCH_LEN);
    strlcpy(s.source, "FIXME",MAX_SOURCE_LEN);

    if (stlisten_verbose > 0)
    {
      fprintf(stderr,"\"Installing\" Appserver 8.1\n");
    }
    rc = st_add(&s, st_env);
    if (stlisten_verbose > 0)
    {
      fprintf(stderr, "return code from add: %d\n", rc);
    }

    st_free_instance_urns(instUUIDs);
  }
}

/*
 * Inits the local agent's product registry for demo purposes.
 *
 * Will only init the product registry if this path:
 *
 *   /opt/SUNWservicetags/bin/initDemo
 *
 * exists.
 */
static void
handle_demo_init(int sockfd)
{
  struct stat buf;
  int i;
  /* note that this method only exists for demo purposes
   * it will be removed before moving to production
   */

  /* TODO: should just wipe out the existing product
   * registry file if it exists and then init the registry
   * with some entries -- instead of trying to call these scripts
   * if they exist
   */

  /* test if the script exists, if so, run it */
  i = stat("/opt/SUNWservicetags/bin/initDemo", &buf);
  if (i == 0)
  {
    if (stlisten_verbose > 0)
    {
      fprintf(stderr, "InitDemo\n");
    }
    exec_demo_script(
      "/opt/SUNWservicetags/bin/initDemo",
      "initDemo",
      "/");
    close(sockfd);
    return;
  }
}

static void
exec_demo_script(char *scriptPath, char *script, char *arg)
{
  pid_t pid;

  /* note that this method only exists for demo purposes
   * it will be removed before moving to production
   */

  pid = fork();
  /* TODO: check that the script is a known script */
  switch (pid)
  {
    case -1:
      if (stlisten_verbose > 0)
      {
        perror("fork");  /* something went wrong */
      }
      exit(1);         /* parent exits */
    case 0:
      if (stlisten_verbose > 0)
      {
        fprintf(stderr, "Executing %s\n", script);
      }
      execl(scriptPath, script, arg, NULL);

      /* will never get here unless we fail the execl */
      exit(0);
    default:
      /* wait for the child to complete */
      wait(NULL);
  }
  return;
}
/* END code for demo */

#endif

/*
static void
handle_get(int sockfd, st_env_t *st_env)
{
  char *buf = NULL;
  int rval;
  char agent_xml[MAX_AGENT_XML];

  if ((rval =  st_get_registry_xml(&buf, st_env)) != ST_ERR_NO_ERROR)
  {
    if (stlisten_verbose > 0)
    {
      fprintf(stderr, "st_get_registry_xml error: %d\n", rval);
    }
    return;
  }

  get_agent_xml(agent_xml);
  sock_printf(sockfd, DO_ENCRYPT, "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n");
  sock_printf(sockfd,
              DO_ENCRYPT, "<st1:response xmlns:st1=\"http://www.sun.com/stv1/svctag\">\n%s%s</st1:response>\n",
              agent_xml,
              &buf[38]);
  st_free_xml(buf);
}
*/

static void
handle_http_get_agent_version(int sockfd)
{
  /*char agent_xml[2048];*/

  /* HTTP Header */
  sock_printf(sockfd, DO_NOT_ENCRYPT, "HTTP/1.0 200 OK\r\n");
  sock_printf(sockfd, DO_NOT_ENCRYPT, "Content-type: text/xml;charset=UTF-8\r\n");
  sock_printf(sockfd, DO_NOT_ENCRYPT, "Connection: Close\r\n");
  sock_printf(sockfd, DO_NOT_ENCRYPT, "\r\n");

  /* HTTP Body */
  sock_printf(sockfd, DO_ENCRYPT, "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n");
  sock_printf(sockfd, DO_ENCRYPT, "<st1:response xmlns:st1=\"http://www.sun.com/stv1/svctag\">\n");
  sock_printf(sockfd, DO_ENCRYPT, " <agent_version>%s</agent_version>\n", stlisten_version);
  sock_printf(sockfd, DO_ENCRYPT, "</st1:response>\n");
}

static void
handle_http_get_agent(int sockfd)
{
  char agent_xml[2048];

  get_agent_xml(agent_xml);

  /* HTTP Header */
  sock_printf(sockfd, DO_NOT_ENCRYPT, "HTTP/1.0 200 OK\r\n");
  sock_printf(sockfd, DO_NOT_ENCRYPT, "Content-type: text/xml;charset=UTF-8\r\n");
  sock_printf(sockfd, DO_NOT_ENCRYPT, "Connection: Close\r\n");
  sock_printf(sockfd, DO_NOT_ENCRYPT, "\r\n");

  /* HTTP Body */
  sock_printf(sockfd, DO_ENCRYPT, "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n");
  sock_printf(sockfd, DO_ENCRYPT, "<st1:response xmlns:st1=\"http://www.sun.com/stv1/svctag\">\n");
  sock_printf(sockfd, DO_ENCRYPT, "%s\n", agent_xml);
  sock_printf(sockfd, DO_ENCRYPT, "</st1:response>\n");
}

/*
 * Note: this function will mangle the contents of first_line.
 */
static void
handle_http_get_svctag(int sockfd, st_env_t *st_env, char *first_line)
{
  st_t svctag;
  int max_encode_len = 4096;
  char encoded[max_encode_len];

  /* TODO: is this the best way to handle max length? can we utilize the fact
   * this value is already set by #define in sthelper.h? */
  sscanf(first_line, "GET /stv1/svctag/%256s HTTP/", svctag.instance_urn);
  st_get(&svctag, st_env);

  /* HTTP Header */
  sock_printf(sockfd, DO_NOT_ENCRYPT, "HTTP/1.0 200 OK\r\n");
  sock_printf(sockfd, DO_NOT_ENCRYPT, "Content-type: text/xml;charset=UTF-8\r\n");
  sock_printf(sockfd, DO_NOT_ENCRYPT, "Connection: Close\r\n");
  sock_printf(sockfd, DO_NOT_ENCRYPT, "\r\n");

  /* HTTP Body */


  sock_printf(sockfd, DO_ENCRYPT, "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n");
  sock_printf(sockfd, DO_ENCRYPT, "<st1:response xmlns:st1=\"http://www.sun.com/stv1/svctag\">\n");
  sock_printf(sockfd, DO_ENCRYPT, " <service_tag>\n");
  sock_printf(sockfd, DO_ENCRYPT, "  <instance_urn>%s</instance_urn>\n", svctag.instance_urn);

  encode_chars((const char *) svctag.product_name, encoded, max_encode_len);
  sock_printf(sockfd, DO_ENCRYPT, "  <product_name>%s</product_name>\n", encoded);

  encode_chars((const char *) svctag.product_version, encoded, max_encode_len);
  sock_printf(sockfd, DO_ENCRYPT, "  <product_version>%s</product_version>\n", encoded);

  sock_printf(sockfd, DO_ENCRYPT, "  <product_urn>%s</product_urn>\n", svctag.product_urn);
  sock_printf(sockfd, DO_ENCRYPT, "  <product_parent_urn>%s</product_parent_urn>\n", svctag.product_parent_urn);

  encode_chars((const char *) svctag.product_parent, encoded, max_encode_len);
  sock_printf(sockfd, DO_ENCRYPT, "  <product_parent>%s</product_parent>\n", encoded);

  encode_chars((const char *) svctag.product_defined_inst_id, encoded, max_encode_len);
  sock_printf(sockfd, DO_ENCRYPT, "  <product_defined_inst_id>%s</product_defined_inst_id>\n", encoded);
  sock_printf(sockfd, DO_ENCRYPT, "  <timestamp>%s</timestamp>\n", svctag.timestamp);

  encode_chars((const char *) svctag.container, encoded, max_encode_len);
  sock_printf(sockfd, DO_ENCRYPT, "  <container>%s</container>\n", encoded);

  encode_chars((const char *) svctag.source, encoded, max_encode_len);
  sock_printf(sockfd, DO_ENCRYPT, "  <source>%s</source>\n", encoded);

  sock_printf(sockfd, DO_ENCRYPT, " </service_tag>\n");
  sock_printf(sockfd, DO_ENCRYPT, "</st1:response>\n");
}

static void
encode_chars(const char *in, char *output, int max_encode_len)
{
    int inlen = strlen(in);
    char *array_ptr = (char *) in;
    int index = 0;
    int loop_cnt = 0;

    while ((*array_ptr) != 0)
    {
        if (*array_ptr == '<')
        {
            if (index+4 >= max_encode_len)
            {
                break;
            }
            output[index++] = '&';
            output[index++] = 'l';
            output[index++] = 't';
            output[index++] = ';';
        }
        else if (*array_ptr == '>')
        {
            if (index+4 >= max_encode_len)
            {
                break;
            }
            output[index++] = '&';
            output[index++] = 'g';
            output[index++] = 't';
            output[index++] = ';';
        }
        else if (*array_ptr == '&')
        {
            /* check if this is already part of an encoding */
            if (loop_cnt+3 < inlen &&
                *(array_ptr+1) == 'l' &&
                *(array_ptr+2) == 't' &&
                *(array_ptr+3) == ';')
            {
                if (index+1 >= max_encode_len)
                {
                    break;
                }
                output[index++] = '&';
            }
            else if (loop_cnt+3 < inlen &&
                     *(array_ptr+1) == 'g' &&
                     *(array_ptr+2) == 't' &&
                     *(array_ptr+3) == ';')
            {
                if (index+1 >= max_encode_len)
                {
                    break;
                }
                output[index++] = '&';
            }
            else if (loop_cnt+4 < inlen &&
                     *(array_ptr+1) == 'a' &&
                     *(array_ptr+2) == 'm' &&
                     *(array_ptr+3) == 'p' &&
                     *(array_ptr+4) == ';')
            {
                if (index+1 >= max_encode_len)
                {
                    break;
                }
                output[index++] = '&';
            }
            else
            {
                if (index+5 >= max_encode_len)
                {
                    break;
                }
                output[index++] = '&';
                output[index++] = 'a';
                output[index++] = 'm';
                output[index++] = 'p';
                output[index++] = ';';
            }
        }
        else
        {
            if (index+1 >= max_encode_len)
            {
                break;
            }
            output[index++] = *array_ptr;
        }
        ++array_ptr;
        loop_cnt++;
        if (index >= max_encode_len)
        {
            break;
        }
    }
    output[index] = '\0';
}

static void
handle_http_get_svctag_list(int sockfd, st_env_t *st_env)
{
  xmlDocPtr registry_doc;
  xmlNodePtr root_node;
  xmlNodePtr node;
  xmlNodePtr node2;
  xmlChar *val;
  int rval;

  if ((rval =  st_get_registry_doc(&registry_doc, st_env)) != ST_ERR_NO_ERROR)
  {
    if (stlisten_verbose > 0)
    {
      fprintf(stderr, "st_get_registry_doc error: %d\n", rval);
    }
    goto DONE;
  }

  if ((root_node = xmlDocGetRootElement(registry_doc)) == NULL)
  {
    if (stlisten_verbose > 0)
    {
      fprintf(stderr, "NULL root node for registry\n");
    }
    goto DONE;
  }

  /* HTTP Header */
  sock_printf(sockfd, DO_NOT_ENCRYPT, "HTTP/1.0 200 OK\r\n");
  sock_printf(sockfd, DO_NOT_ENCRYPT, "Content-type: text/xml;charset=UTF-8\r\n");
  sock_printf(sockfd, DO_NOT_ENCRYPT, "Connection: Close\r\n");
  sock_printf(sockfd, DO_NOT_ENCRYPT, "\r\n");

  /* HTTP Body */
  sock_printf(sockfd, DO_ENCRYPT, "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n");
  sock_printf(sockfd, DO_ENCRYPT, "<?xml-stylesheet type=\"text/xsl\"");
  sock_printf(sockfd, DO_ENCRYPT, " href=\"/stv1/st_list.xsl\"");
  sock_printf(sockfd, DO_ENCRYPT, " media=\"screen\"?>\n");
  sock_printf(sockfd, DO_ENCRYPT, "<service_tags>\n");
  for (node = root_node->xmlChildrenNode; node != NULL; node = node->next)
  {
    if (xmlStrcmp(node->name, BAD_CAST "service_tag") == 0)
    {
      for (node2 = node->xmlChildrenNode; node2 != NULL; node2 = node2->next)
      {
        if (xmlStrcmp(node2->name, BAD_CAST "instance_urn") == 0)
        {
          val = xmlNodeListGetString(registry_doc, node2->xmlChildrenNode, 1);
          if (val != NULL)
          {
            sock_printf(sockfd,
                        DO_ENCRYPT, "  <link type=\"service_tag\" href=\"/stv1/svctag/%s\" />\n",
                        val);
            xmlFree(val);
          }
        }
      }
    }
  }
  sock_printf(sockfd, DO_ENCRYPT, "</service_tags>\n");

DONE:
  if (registry_doc != NULL)
  {
    xmlFreeDoc(registry_doc);
  }
}

static void
handle_http_get_svctag_list_xsl(int sockfd)
{
  /* HTTP Header */
  sock_printf(sockfd, DO_NOT_ENCRYPT, "HTTP/1.0 200 OK\r\n");
  sock_printf(sockfd, DO_NOT_ENCRYPT, "Content-type: text/xml;charset=UTF-8\r\n");
  sock_printf(sockfd, DO_NOT_ENCRYPT, "Connection: Close\r\n");
  sock_printf(sockfd, DO_NOT_ENCRYPT, "\r\n");

  /* HTTP Body */
  sock_printf(sockfd, DO_ENCRYPT, "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n");
  sock_printf(sockfd, DO_ENCRYPT, "<html xmlns:xsl=\"http://www.w3.org/1999/XSL/Transform\" xsl:version=\"1.0\">\n");
  sock_printf(sockfd, DO_ENCRYPT, "  <body>\n");
  sock_printf(sockfd, DO_ENCRYPT, "    <table border=\"2\" bgcolor=\"yellow\">\n");
  sock_printf(sockfd, DO_ENCRYPT, "      <tr>\n");
  sock_printf(sockfd, DO_ENCRYPT, "        <th>Service Tags Exposed By This Agent</th>\n");
  sock_printf(sockfd, DO_ENCRYPT, "      </tr>\n");
  sock_printf(sockfd, DO_ENCRYPT, "      <xsl:for-each select=\"service_tags/link\">\n");
  sock_printf(sockfd, DO_ENCRYPT, "      <tr>\n");
  sock_printf(sockfd, DO_ENCRYPT, "        <td>\n");
  sock_printf(sockfd, DO_ENCRYPT, "          <a>\n");
  sock_printf(sockfd, DO_ENCRYPT, "            <xsl:attribute name=\"href\">\n");
  sock_printf(sockfd, DO_ENCRYPT, "              <xsl:value-of select=\"@href\" />\n");
  sock_printf(sockfd, DO_ENCRYPT, "            </xsl:attribute>\n");
  sock_printf(sockfd, DO_ENCRYPT, "            <xsl:value-of select=\"@href\" />\n");
  sock_printf(sockfd, DO_ENCRYPT, "          </a>\n");
  sock_printf(sockfd, DO_ENCRYPT, "        </td>\n");
  sock_printf(sockfd, DO_ENCRYPT, "      </tr>\n");
  sock_printf(sockfd, DO_ENCRYPT, "      </xsl:for-each>\n");
  sock_printf(sockfd, DO_ENCRYPT, "    </table>\n");
  sock_printf(sockfd, DO_ENCRYPT, "  </body>\n");
  sock_printf(sockfd, DO_ENCRYPT, "</html>\n");
}

void
get_agent_xml(char *agent_xml)
{
  int i;
  char tmp[512];
  char urn[512];
  char a_version[512];
  char r_version[512];

  struct probeStruct probeData;

  strlcpy(agent_xml, "<agent>\n", MAX_AGENT_XML);
  snprintf(urn, 512, "  <agent_urn>%s</agent_urn>\n", st_get_registry_urn());

  snprintf(a_version, 512,
           "  <agent_version>%s</agent_version>\n",
           stlisten_version);

  snprintf(r_version, 512,
           "  <registry_version>%s</registry_version>\n",
           registry_version);

  strlcat(agent_xml, urn, MAX_AGENT_XML);
  strlcat(agent_xml, a_version, MAX_AGENT_XML);
  strlcat(agent_xml, r_version, MAX_AGENT_XML);
  strlcat(agent_xml, "  <system_info>\n", MAX_AGENT_XML);

  get_probe_data(&probeData);
  for (i = 0; i < MAX_SYSCONF_FIELDS; i++)
  {
    /*
    sprintf(tmp,
      "    <%s>%s</%s>\n",
      sysconfField[i],
      probeData.sysconf[i],
      sysconfField[i]);
      */
    snprintf(tmp,
      512,
      "    <%s>%s</%s>\n",
      sysconfField[i],
      probeData.sysconf[i],
      sysconfField[i]);
    strlcat(agent_xml, tmp, MAX_AGENT_XML);
  }
  strlcat(agent_xml, "  </system_info>\n", MAX_AGENT_XML);

  strlcat(agent_xml, "</agent>", MAX_AGENT_XML);
}

/*
 * Prints output to a socket, following printf conventions for format codes,
 * etc.
 *
 * encrypt_buf is optional and is only used if STCRYPTO_UEF or STCRYPTO_EVP
 * is defined.  Some buffers should not be encrypted, even if we're running
 * with encryption on.  Some examples would include our HTTP headers and
 * status codes.
 */
static int
sock_printf(int sockfd, int encrypt_buf, char *format, ...)
{
  va_list ap;
  char msgbuf[BUF_LEN];
  #ifdef STCRYPTO
  int len;
  unsigned char encoded_buf[BUF_LEN];
  unsigned char encrypted_buf[BUF_LEN];
  #endif

  va_start(ap, format);
  if (vsnprintf(msgbuf, BUF_LEN, format, ap) < 0)
  {
    if (stlisten_verbose > 0)
    {
      perror("vsprintf");
    }
    va_end(ap);
    return -1;
  }
  va_end(ap);

  if (stlisten_verbose > 1)
  {
    fprintf(stderr, "%s\n", msgbuf);
  }

  #ifdef STCRYPTO
  if (encrypt_buf && do_encrypt)
  {
    #ifdef STCRYPTO_EVP
    len = aes_evp_encrypt((unsigned char *) encryption_key,
                    (unsigned const char *) msgbuf,
                    strlen(msgbuf),
                    encrypted_buf);
    #endif

    #ifdef STCRYPTO_UEF
    len = aes_uef_encrypt((unsigned char *) encryption_key,
                    (unsigned const char *) msgbuf,
                    strlen(msgbuf),
                    encrypted_buf);
    #endif

    b64_ntop((u_char const *)encrypted_buf, len, (char *) encoded_buf,
      BUF_LEN);

    /* append newline to end of encodedbuf, receiver is responsible */
    /* for stripping this off before decoding & decrypting */
    strlcat((char *) encoded_buf, "\n", BUF_LEN);

    if (send(sockfd, encoded_buf, strlen((char*)encoded_buf), 0) < 0)
    {
      if (stlisten_verbose > 0)
      {
        perror("send");
      }
      va_end(ap);
      return -1;
    }
  }
  else
  #endif
  {
    if (send(sockfd, msgbuf, strlen(msgbuf), 0) < 0)
    {
      if (stlisten_verbose > 0)
      {
        perror("send");
      }
      va_end(ap);
      return -1;
    }
  }

  return 0;
}

/*
 * Get one line of input from an open socket.  Returns NULL if the socket
 * is closed or immediately reaches EOF prior to any data.
 */
static char *
sock_gets(int sockfd, char *s, int n)
{
  int numbytes;
  int i;

  for (i = 0; i < n; i++)
  {
    numbytes = read(sockfd, &(s[i]), 1);
    if (numbytes < 0)
    {
      if (stlisten_verbose > 0)
      {
        perror("read");
      }
      return NULL;
    }
    else if ((numbytes == 0) || (s[i] == '\n'))
    {
      s[i] = '\0';

      if ((i == 0) && (numbytes == 0))
      {
        return NULL;
      }
      else
      {
        if (stlisten_verbose > 1)
        {
          fprintf(stderr, "sock_gets: %s\n", s);
        }
        return s;
      }
    }
  }

  if (stlisten_verbose > 1)
  {
    fprintf(stderr,
          "warning: sock_gets reached end of buffer, truncation likely\n");
  }
  s[n-1] = '\0';
  return s;
}

static char
*sock_ntop(const struct sockaddr *sa, socklen_t salen)
{
  char portstr[7];
  static char str[128];

  switch (sa->sa_family)
  {
    case AF_INET:
    {
      struct sockaddr_in *sin = (struct sockaddr_in *) sa;
      if (inet_ntop(AF_INET, &sin->sin_addr, str, sizeof(str)) == NULL)
      {
        return (NULL);
      }
      if (ntohs(sin->sin_port) != 0)
      {
        snprintf(portstr, sizeof(portstr), ".%d", ntohs(sin->sin_port));
        strlcat(str, portstr, 128);
      }

      return (str);
    }
    case AF_INET6:
    {
      struct sockaddr_in6     *sin6 = (struct sockaddr_in6 *) sa;

      if (inet_ntop(AF_INET6, &sin6->sin6_addr, str, sizeof(str)) == NULL)
                        return(NULL);
      if (ntohs(sin6->sin6_port) != 0) {
                        snprintf(portstr, sizeof(portstr), ".%d", ntohs(sin6->sin6_port));
                        strlcat(str, portstr, 128);
      }
      return(str);
    }
  }
  return (NULL);
}

#ifdef LINUX
static void
read_and_discard_data(int sockfd)
{
  int rc;
  struct pollfd pollfd;
  int timeout = 125;
  char buf[PATH_MAX];

   memset(&pollfd, 0, sizeof pollfd);
   pollfd.fd = sockfd;
   pollfd.events = POLLIN;

   while ((rc = poll(&pollfd, 1, timeout)) > 0) {

       if (pollfd.revents & POLLHUP) {
           return;
       }
       if (pollfd.revents & POLLIN) {
           (void)sock_gets(sockfd, buf, sizeof (buf));
       } else {
           return;
       }

       memset(&pollfd, 0, sizeof pollfd);
       pollfd.fd = sockfd;
       pollfd.events = POLLIN;
   }
}
#endif

