/*
 * Copyright 2008 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

/*
 * Responds to stcollector requests for information.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <signal.h>
#include <pthread.h>
#include <pwd.h>
#include <grp.h>
#include <libgen.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <syslog.h>
#ifdef SMF
#include <libscf.h>
#include "scf.h"
#endif

#include "sthelper.h"
#include "stlisten.h"
#include "probe.h"
#include "comm.h"

#include "strlcpy.h"

/* The default TCP port on which we'll listen for communication requests. */
#define DEFAULT_TCP_PORT 6481

/* Size of the buffer we'll use for various purposes. */
#define BUF_LEN (128 * 1024)

#ifdef LINUX
#define MAXNAMELEN 256
#endif

#define MAXKEYLEN 16
/* use KEYBUFLEN for the buffer size, 2 larger than the keysize because
 * of the need to include the \n and \0 chars in the buffer */
#define KEYBUFLEN (MAXKEYLEN + 2)
#define MINKEYLEN 5

/*static void * launch_tcp_listener(void *);*/
static void make_daemon_process();
static void set_uid(char *);
static void set_gid(char *);
static int  process_passphrase(char *, char *);

void retrieve_solaris_system_info(char *, char *, char*, char *);

/* Setting this to a higher value increases diagnostic output. */
int stlisten_verbose = 0;

/* listener version */
char stlisten_version[] = SERVICE_TAG_VERSION;

/* registry version */
char registry_version[512];

/* Agent context information */
agent_ctx_t agent_ctx;

/* System environmental information */
typedef struct st_sysinfo {
    char psn[MAXNAMELEN];         /* product serial number */
    char cpuman[MAXNAMELEN];      /* CPU manufacturer */
    char sysman[MAXNAMELEN];      /* system manufactuer */
    char prodname[MAXNAMELEN];    /* product name (x86) */
} st_sysinfo_t;

static st_sysinfo_t st_sysinfo;

/* encryption values */
char encryption_key[KEYBUFLEN];
int do_encrypt = 0;

#ifdef SMF
/*
 * Table of permitted SMF properties that can be used to override
 * default values.  Used for SMF (inetd) control only.
 */
scf_param_t smfprops[] =
{
    { SCF_ST_DISCOVERY_REGISTRY,           "registry_location",  NULL },
    { SCF_ST_DISCOVERY_PASSPHRASE ,        "passphrase",         NULL },
    { SCF_ST_DISCOVERY_LOCAL_SUBNET_ONLY , "local_subnet_only",  NULL },
    { SCF_ST_DISCOVERY_LOCAL_LOOPBACK_ONLY , "local_loopback_only",  NULL },
    { SCF_ST_DISCOVERY_NULL,                NULL,                NULL }
};
#endif

/*
 * Program invocation entry point.
 */
int main(int argc, char **argv)
{
  /*pthread_t sn_thread;*/
  int c;
  int rc;
  int errors = 0;
  int do_daemon = 1;
  int local_subnet_only = 0;
  int local_loopback_only = 0;
  char *registry_location = NULL;
  char *uid = NULL;
  char *gid = NULL;
  int tcp_port = DEFAULT_TCP_PORT;
  extern char *optarg;
  extern int optopt;
  int rval;
  int rlen;
  /*int klen;*/
 
  /* Used to determine if we should be running under inetd control or not. */
  /* Default is to not run under inetd control */
  static int inetd = 0;

  /* Service registry tag state information */
  static st_env_t st_env;

  while ((c = getopt(argc, argv, "fvisL:p:r:t:u:D:N:U:G:l:")) != -1)
  {
    switch (c)
    {
      case 'f':
        do_daemon = 0;
        break;
      case 'L':
        local_loopback_only = 1;
        break;
      case 'i':
        inetd = 1;
        break;
      case 's':
        local_subnet_only = 1;
        break;
      case 'v':
        stlisten_verbose++;
        break;
      case 'p':
        #ifndef STCRYPTO
        /* is this how we want to handle this situation? */
        fprintf(stderr,
                "Encryption support not enabled, ignoring encryption key\n");
        do_encrypt = 0;
        if (do_encrypt == 0)
        {
          break;
        }
        #endif

        rc = process_passphrase(optarg, encryption_key);
        if (rc == -1)
        {
          fprintf(stderr,
            "-p passphrase contained in file must be between %d and %d characters long\n",
            MINKEYLEN, MAXKEYLEN);
          return 1;
        }
        else if (rc == -2)
        {
          fprintf(stderr,
            "-p passphrase file must exist or be readable\n");
          return 1;
        }
        else if (rc != 0)
        {
          fprintf(stderr,
            "-p passphrase unknown error\n");
          return 1;
        }
        do_encrypt = 1;
        break;
      case 'r':
        rlen = strlen(optarg) + 1;
        if ((registry_location = (char *)malloc(rlen)) == NULL)
        {
          perror("malloc(registry_location)");
          return 1;
        }
        if (strlcpy(registry_location, optarg, rlen) >= rlen)
        {
          fprintf(stderr, "-r registry location path too long\n");
          return 1;
        }
        break;
      case 't':
        tcp_port = atoi(optarg);
        break;
      case 'U':
        rlen = strlen(optarg) + 1;
        if ((uid = (char *)malloc(rlen)) == NULL)
        {
          perror("malloc(uid)");
          return 1;
        }
        if (strlcpy(uid, optarg, rlen) >= rlen)
        {
          fprintf(stderr, "-U uid name too long\n");
          return 1;
        }
        break;
      case 'G':
        rlen = strlen(optarg) + 1;
        if ((gid = (char *)malloc(rlen)) == NULL)
        {
          perror("malloc(gid)");
          return 1;
        }
        if (strlcpy(gid, optarg, rlen) >= rlen)
        {
          fprintf(stderr, "-g gid name too long\n");
          return 1;
        }
        break;
      case 'l':
        if ((rval = open(optarg,(O_RDWR|O_APPEND|O_CREAT),0666)) > 0)
        {
          (void) dup2(rval, 2);
          (void) close(rval);
        }
        else
        {
          /* Non-fatal error */
          fprintf(stderr, "Unable to open log file: %s\n", optarg);
        }
        break;
      case ':':
        fprintf(stderr, "Option -%c requires an operand\n", optopt);
        errors++;
        break;
      case '?':
        fprintf(stderr, "Unrecognized option: -%c\n", optopt);
        errors++;
        break;
    }
  }

  if (errors)
  {
    fprintf(stderr, "Usage: %s [options]\n", basename(argv[0]));
    fprintf(stderr, "Options:\n");
    fprintf(stderr, "  -f                     do not fork() from the calling shell\n");
    fprintf(stderr, "  -l logfile             redirect stderr to specified file\n");
    fprintf(stderr, "  -i                     run under control of inetd(1M)\n");
    fprintf(stderr, "  -L                     only respond to requests from the local loopback address\n");
    fprintf(stderr, "  -s                     only respond to requests in the local subnet\n");
    fprintf(stderr, "  -v                     provide verbose output\n");
    fprintf(stderr, "  -p password_file       use the password contained in the file (must be 5 to 16 characters long) for AES encryption\n");
    fprintf(stderr, "  -r alt_root_dir        look for Service Tag Registry under specified dir\n");
    fprintf(stderr, "  -t tcp_port            use the specified udp port for communication\n");
    fprintf(stderr, "  -U user                setuid to specified user\n");
    fprintf(stderr, "  -G group               setgid to specified group\n");

    return errors;
  }

  if (gid != NULL)
  {
    set_gid(gid);
  }
  if (uid != NULL)
  {
    set_uid(uid);
  }

  memset(&st_sysinfo, 0, sizeof st_sysinfo);
#ifdef SOLARIS_PSN
  /*
   * Retrieving host serial number, CPU manufacturer, and system
   * manufacturer takes greater than basic privileges.  Grab this
   * information now so that we can reduce privileges immediately
   * after.
   */

   retrieve_solaris_system_info(st_sysinfo.psn, st_sysinfo.cpuman,
                                 st_sysinfo.sysman,
                                 st_sysinfo.prodname);
   if (stlisten_verbose) {
      fprintf(stderr, "CSN = %s\n", st_sysinfo.psn);
      fprintf(stderr, "Man = %s\n", st_sysinfo.sysman);
      fprintf(stderr, "CPU Man = %s\n", st_sysinfo.cpuman);
      fprintf(stderr, "Product name = %s\n", st_sysinfo.prodname);
   }
#endif

  /*
   * In an inetd environment, the UDP socket is passed via fd zero.  No
   * need to daemonize in this mode.
   */
  if (inetd)
  {
    do_daemon = 0;
    /* if we're in inetd, stderr will print to the socket, which we don't
       want, so, we'll set stlisten_verbose to 0, and wrap all fprintf(stderr,...
       messages with "if (stlisten_verbose)"
    */
    stlisten_verbose = 0;
#ifdef SMF
    /*
     *  In inetd-invocations, the published SMF properties
     *  (if present) override any existing values.
     */

      process_smf_properties(smfprops);

      for (int i = 0; smfprops[i].id != SCF_ST_DISCOVERY_NULL; i++)
      {
          char *propval = smfprops[i].value;
          if (propval != NULL)
          {
              switch (smfprops[i].id)
              {
                 case SCF_ST_DISCOVERY_REGISTRY:
                    if (strlen(propval) > 0)
                    {
                       if (registry_location) free(registry_location);
                       rlen = strlen(propval) + 1;
                       if ((registry_location = (char *)malloc(rlen)) != NULL)
                       {
                          strlcpy(registry_location, propval, rlen);
                       }
                    }
                    break;
                 case SCF_ST_DISCOVERY_PASSPHRASE:
                    if (strlen(propval) == 0)
                    {
                        do_encrypt = 0;
                    }
                    else if (process_passphrase(propval, encryption_key) == 0)
                    {
                        do_encrypt = 1;
                    }
                    break;
                 case SCF_ST_DISCOVERY_LOCAL_SUBNET_ONLY:
                    if (strncasecmp(propval, SMF_BOOLEAN_TRUE,
                        strlen(SMF_BOOLEAN_TRUE)) == 0)
                    {
                        local_subnet_only = 1;
                    }
                    else if (strncasecmp(propval, SMF_BOOLEAN_FALSE,
                            strlen(SMF_BOOLEAN_FALSE)) == 0)
                    {
                        local_subnet_only = 0;
                    }
                    break;
                 case SCF_ST_DISCOVERY_LOCAL_LOOPBACK_ONLY:
                    if (strncasecmp(propval, SMF_BOOLEAN_TRUE,
                            strlen(SMF_BOOLEAN_TRUE)) == 0)
                    {
                        local_loopback_only = 1;
                    }
                    else if (strncasecmp(propval, SMF_BOOLEAN_FALSE,
                            strlen(SMF_BOOLEAN_FALSE)) == 0)
                    {
                        local_loopback_only = 0;
                    }
                    break;
                 default:
                    break;
               }
          }
      }
#endif
  }
  else
  {
    /* We shouldn't die just because collector closes its end of the socket. */
    /* (TODO: should probably have our own custom handler) */
    signal(SIGPIPE, SIG_IGN);
  }

  if (do_daemon)
  {
    make_daemon_process();
  }

  memset(&st_env, 0, sizeof st_env);
  if (st_initialize(registry_location, &st_env) < 0)
  {
    free(registry_location);
  }

  strlcpy(registry_version, st_env.registry_version, 512);

  memset(&agent_ctx, 0, sizeof agent_ctx);
  agent_ctx.port   = tcp_port;
  agent_ctx.st_env = &st_env;

  receive_tcp_connections(inetd,
                          local_subnet_only,
                          local_loopback_only,
                          agent_ctx.port,
                          agent_ctx.st_env);

  return 0;
}

static void
make_daemon_process()
{
  /* BEGIN - attempt to make this a daemon */
  /* Ignore terminal I/O signals */
  signal(SIGTTOU, SIG_IGN);
  signal(SIGTTIN, SIG_IGN);
  signal(SIGTSTP, SIG_IGN);
  /* Ignore terminal I/O signals */

  int pid;
  pid = fork();
  if (pid != 0) {
    exit(0); /* this is the parent process here, so let's exit it for our daemon work */
  }

  setsid(); /* child process, so let's make us a session leader, and lose control tty */
  signal(SIGHUP, SIG_IGN); /* SIGHUP will be sent to child process so ignore it */

  if (pid != 0) {
    exit(0);
  }
  /* END - attempt to make this a daemon */
}

/* Attempts to change the uid and gid of the program to the svctag user
 * and the svctag's group (which should be svctag) */
static void
set_uid(char *user_name)
{
  struct passwd *user;

  if (user_name == NULL)
  {
    if (stlisten_verbose)
    {
      fprintf(stderr, "WARNING: no user specified\n");
    }
    return;
  }

  user = getpwnam(user_name);

  if (user == NULL)
  {
    if (stlisten_verbose)
    {
      fprintf(stderr, "WARNING: no such user exists: %s\n", user_name);
    }
    return;
  }

  /* attempt to set the uid and gid of the program -- only works if
   * the user has appropriate permissions */
  if (setuid(user->pw_uid) != 0)
  {
    perror("setuid");
  }
}


/* Attempts to change the uid and gid of the program to the svctag user
 * and the svctag's group (which should be svctag) */
static void
set_gid(char *group_name)
{
  struct group *group;

  if (group_name == NULL)
  {
    if (stlisten_verbose)
    {
      fprintf(stderr, "WARNING: no group specified\n");
    }
    return;
  }

  group = getgrnam(group_name);

  if (group == NULL)
  {
    if (stlisten_verbose)
    {
      fprintf(stderr, "WARNING: no such group exists: %s\n", group_name);
    }
    return;
  }

  if (setgid(group->gr_gid) != 0)
  {
    perror("setgid");
  }
}

/*
 * Processes the passphrase contained in the source file.  Stores
 * the passphrase in the dest parameter.
 *
 */
static int
process_passphrase(char *source_file, char *dest)
{
   char buffer[KEYBUFLEN];

   /* try to read from the file and store the contents to source */
   /*
   if source_file does not exist or source_file is not readable
   then return -2
   */
   FILE *fp;
   int fd;
   int n;

   fd = open(source_file, O_RDONLY);
   if (fd < 0)
   {
      return -2;
   }
   fp = fdopen(fd, "r");
   if (fp == NULL)
   {
      return -2;
   }

   memset(buffer, '\0', KEYBUFLEN);

   n = fread(buffer, 1, KEYBUFLEN, fp);
   if (n < 0) {
      fclose(fp);
      return -2;
   }

   if (n >= KEYBUFLEN)
   {
      return -1;
   }

   if (buffer != NULL && dest != NULL)
   {
      int klen = strlen(buffer);
      if (klen > KEYBUFLEN
            || klen < (MINKEYLEN+1)
            || (strlcpy(dest, buffer, klen) >= KEYBUFLEN))
      {
         return -1;
      }

      return 0;
   }
   return -3;
}

char *
solaris_psn()
{
    return st_sysinfo.psn;
}

char *
solaris_sysman()
{
    return st_sysinfo.sysman;
}

char *
solaris_cpuman()
{
    return st_sysinfo.cpuman;
}

char *
solaris_prodname()
{
    return st_sysinfo.prodname;
}

